@@ -18,7 +18,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
     
 #include <stdio.h>
 #include <string.h>
@@ -59,7 +60,7 @@ int arm_compare_fp;
 /* What type of cpu are we compiling for? */
 enum processor_type arm_cpu;
 
-/* Waht type of floating point are we compiling for? */
+/* What type of floating point are we compiling for? */
 enum floating_point_type arm_fpu;
 
 /* In case of a PRE_INC, POST_INC, PRE_DEC, POST_DEC memory reference, we
@@ -1143,6 +1144,32 @@ di_operand (op, mode)
     }
 }
 
+/* Return TRUE for a valid operand of a DFmode operation when -msoft-float.
+   Either: REG, CONST_DOUBLE or MEM(DImode_address).
+   Note that this disallows MEM(REG+REG), but allows
+   MEM(PRE/POST_INC/DEC(REG)).  */
+
+int
+soft_df_operand (op, mode)
+     rtx op;
+     enum machine_mode mode;
+{
+  if (s_register_operand (op, mode))
+    return TRUE;
+
+  switch (GET_CODE (op))
+    {
+    case CONST_DOUBLE:
+      return TRUE;
+
+    case MEM:
+      return memory_address_p (DFmode, XEXP (op, 0));
+
+    default:
+      return FALSE;
+    }
+}
+
 /* Return TRUE for valid index operands. */
 
 int
@@ -1472,8 +1499,8 @@ store_multiple_operation (op, mode)
 
 int
 multi_register_push (op, mode)
-  rtx op;
-  enum machine_mode mode;
+     rtx op;
+     enum machine_mode mode;
 {
   if (GET_CODE (op) != PARALLEL
       || (GET_CODE (XVECEXP (op, 0, 0)) != SET)
@@ -1713,6 +1740,35 @@ gen_compare_reg (code, x, y, fp)
   return cc_reg;
 }
 
+void
+arm_reload_in_hi (operands)
+     rtx *operands;
+{
+  rtx base = find_replacement (&XEXP (operands[1], 0));
+
+  emit_insn (gen_zero_extendqisi2 (operands[2], gen_rtx (MEM, QImode, base)));
+  emit_insn (gen_zero_extendqisi2 (gen_rtx (SUBREG, SImode, operands[0], 0),
+				   gen_rtx (MEM, QImode, 
+					    plus_constant (base, 1))));
+  if (BYTES_BIG_ENDIAN)
+    emit_insn (gen_rtx (SET, VOIDmode, gen_rtx (SUBREG, SImode, 
+						operands[0], 0),
+			gen_rtx (IOR, SImode, 
+				 gen_rtx (ASHIFT, SImode,
+					  gen_rtx (SUBREG, SImode,
+						   operands[0], 0),
+					  GEN_INT (8)),
+				 operands[2])));
+  else
+    emit_insn (gen_rtx (SET, VOIDmode, gen_rtx (SUBREG, SImode, 
+						operands[0], 0),
+			gen_rtx (IOR, SImode, 
+				 gen_rtx (ASHIFT, SImode,
+					  operands[2],
+					  GEN_INT (8)),
+				 gen_rtx (SUBREG, SImode, operands[0], 0))));
+}
+
 void
 arm_reload_out_hi (operands)
      rtx *operands;
@@ -1827,14 +1883,14 @@ print_multi_reg (stream, instr, mask, hat)
   int not_first = FALSE;
 
   fputc ('\t', stream);
-  fprintf (stream, instr, ARM_REG_PREFIX);
+  fprintf (stream, instr, REGISTER_PREFIX);
   fputs (", {", stream);
   for (i = 0; i < 16; i++)
     if (mask & (1 << i))
       {
 	if (not_first)
 	  fprintf (stream, ", ");
-	fprintf (stream, "%s%s", ARM_REG_PREFIX, reg_names[i]);
+	fprintf (stream, "%s%s", REGISTER_PREFIX, reg_names[i]);
 	not_first = TRUE;
       }
 
@@ -2370,7 +2426,7 @@ shift_op (op, amountp)
 	 multiplication by a power of 2 with the recognizer for a
 	 shift.  >=32 is not a valid shift for "asl", so we must try and
 	 output a shift that produces the correct arithmetical result.
-	 Using lsr #32 is idendical except for the fact that the carry bit
+	 Using lsr #32 is identical except for the fact that the carry bit
 	 is not set correctly if we set the flags; but we never use the 
 	 carry bit from such an operation, so we can ignore that.  */
       if (code == ROTATERT)
@@ -2690,6 +2746,12 @@ output_return_instruction (operand, really_return)
   return "";
 }
 
+/* Return nonzero if optimizing and the current function is volatile.
+   Such functions never return, and many memory cycles can be saved
+   by not storing register values that will never be needed again.
+   This optimization was added to speed up context switching in a
+   kernel application. */
+
 int
 arm_volatile_func ()
 {
@@ -2734,15 +2796,15 @@ output_func_prologue (f, frame_size)
   return_used_this_function = 0;
   lr_save_eliminated = 0;
   
-  fprintf (f, "\t%c args = %d, pretend = %d, frame = %d\n",
-	   ARM_COMMENT_CHAR, current_function_args_size,
+  fprintf (f, "\t%s args = %d, pretend = %d, frame = %d\n",
+	   ASM_COMMENT_START, current_function_args_size,
 	   current_function_pretend_args_size, frame_size);
-  fprintf (f, "\t%c frame_needed = %d, current_function_anonymous_args = %d\n",
-	   ARM_COMMENT_CHAR, frame_pointer_needed,
+  fprintf (f, "\t%s frame_needed = %d, current_function_anonymous_args = %d\n",
+	   ASM_COMMENT_START, frame_pointer_needed,
 	   current_function_anonymous_args);
 
   if (volatile_func)
-    fprintf (f, "\t%c Volatile function.\n", ARM_COMMENT_CHAR);
+    fprintf (f, "\t%s Volatile function.\n", ASM_COMMENT_START);
 
   if (current_function_anonymous_args && current_function_pretend_args_size)
     store_arg_regs = 1;
@@ -2775,8 +2837,8 @@ output_func_prologue (f, frame_size)
     }
 
   if (lr_save_eliminated)
-    fprintf (f,"\t%c I don't think this function clobbers lr\n",
-	     ARM_COMMENT_CHAR);
+    fprintf (f,"\t%s I don't think this function clobbers lr\n",
+	     ASM_COMMENT_START);
 }
 
 
@@ -2822,8 +2884,8 @@ output_func_epilogue (f, frame_size)
       for (reg = 23; reg > 15; reg--)
 	if (regs_ever_live[reg] && ! call_used_regs[reg])
 	  {
-	    fprintf (f, "\tldfe\t%s%s, [%sfp, #-%d]\n", ARM_REG_PREFIX,
-		     reg_names[reg], ARM_REG_PREFIX, floats_offset);
+	    fprintf (f, "\tldfe\t%s%s, [%sfp, #-%d]\n", REGISTER_PREFIX,
+		     reg_names[reg], REGISTER_PREFIX, floats_offset);
 	    floats_offset += 12;
 	    code_size += 4;
 	  }
@@ -2846,8 +2908,8 @@ output_func_epilogue (f, frame_size)
       for (reg = 16; reg < 24; reg++)
 	if (regs_ever_live[reg] && ! call_used_regs[reg])
 	  {
-	    fprintf (f, "\tldfe\t%s%s, [%ssp], #12\n", ARM_REG_PREFIX,
-		     reg_names[reg], ARM_REG_PREFIX);
+	    fprintf (f, "\tldfe\t%s%s, [%ssp], #12\n", REGISTER_PREFIX,
+		     reg_names[reg], REGISTER_PREFIX);
 	    code_size += 4;
 	  }
       if (current_function_pretend_args_size == 0 && regs_ever_live[14])
@@ -2873,7 +2935,7 @@ output_func_epilogue (f, frame_size)
 	    }
 	  fprintf (f,
 		   TARGET_6 ? "\tmov\t%spc, %slr\n" : "\tmovs\t%spc, %slr\n",
-		   ARM_REG_PREFIX, ARM_REG_PREFIX, f);
+		   REGISTER_PREFIX, REGISTER_PREFIX, f);
 	  code_size += 4;
 	}
     }
@@ -2977,7 +3039,7 @@ arm_expand_prologue ()
   if (live_regs_mask)
     {
       /* If we have to push any regs, then we must push lr as well, or
-	 we won't get a propper return.  */
+	 we won't get a proper return.  */
       live_regs_mask |= 0x4000;
       emit_multi_reg_push (live_regs_mask);
     }
@@ -3014,13 +3076,13 @@ arm_expand_prologue ()
 
 /* If CODE is 'd', then the X is a condition operand and the instruction
    should only be executed if the condition is true.
-   if CODE is 'D', then the X is a condition operand and the instruciton
+   if CODE is 'D', then the X is a condition operand and the instruction
    should only be executed if the condition is false: however, if the mode
    of the comparison is CCFPEmode, then always execute the instruction -- we
    do this because in these circumstances !GE does not necessarily imply LT;
    in these cases the instruction pattern will take care to make sure that
    an instruction containing %d will follow, thereby undoing the effects of
-   doing this instrucion unconditionally.
+   doing this instruction unconditionally.
    If CODE is 'N' then X is a floating point operand that must be negated
    before output.
    If CODE is 'B' then output a bitwise inverted value of X (a const int).
@@ -3035,11 +3097,11 @@ arm_print_operand (stream, x, code)
   switch (code)
     {
     case '@':
-      fputc (ARM_COMMENT_CHAR, stream);
+      fputs (ASM_COMMENT_START, stream);
       return;
 
     case '|':
-      fputs (ARM_REG_PREFIX, stream);
+      fputs (REGISTER_PREFIX, stream);
       return;
 
     case '?':
@@ -3105,12 +3167,12 @@ arm_print_operand (stream, x, code)
     case 'R':
       if (REGNO (x) > 15)
 	abort ();
-      fputs (ARM_REG_PREFIX, stream);
+      fputs (REGISTER_PREFIX, stream);
       fputs (reg_names[REGNO (x) + 1], stream);
       return;
 
     case 'm':
-      fputs (ARM_REG_PREFIX, stream);
+      fputs (REGISTER_PREFIX, stream);
       if (GET_CODE (XEXP (x, 0)) == REG)
 	fputs (reg_names[REGNO (XEXP (x, 0))], stream);
       else
@@ -3118,8 +3180,8 @@ arm_print_operand (stream, x, code)
       return;
 
     case 'M':
-      fprintf (stream, "{%s%s-%s%s}", ARM_REG_PREFIX, reg_names[REGNO (x)],
-	       ARM_REG_PREFIX, reg_names[REGNO (x) - 1
+      fprintf (stream, "{%s%s-%s%s}", REGISTER_PREFIX, reg_names[REGNO (x)],
+	       REGISTER_PREFIX, reg_names[REGNO (x) - 1
 					 + ((GET_MODE_SIZE (GET_MODE (x))
 					     + GET_MODE_SIZE (SImode) - 1)
 					    / GET_MODE_SIZE (SImode))]);
@@ -3148,7 +3210,7 @@ arm_print_operand (stream, x, code)
 
       if (GET_CODE (x) == REG)
 	{
-	  fputs (ARM_REG_PREFIX, stream);
+	  fputs (REGISTER_PREFIX, stream);
 	  fputs (reg_names[REGNO (x)], stream);
 	}
       else if (GET_CODE (x) == MEM)
@@ -3208,7 +3270,7 @@ arm_asm_output_label (stream, name)
   else
     {
       real_name = xmalloc (2 + strlen (name));
-      strcpy (real_name, "_");
+      strcpy (real_name, USER_LABEL_PREFIX);
       strcat (real_name, name);
     }
   for (s = real_name; *s; s++)
@@ -3259,7 +3321,7 @@ output_load_symbol (insn, operands)
     abort ();
 
   /* When generating the instructions, we never mask out the bits that we
-     think will be always zero, then if a mistake has occured somewhere, the
+     think will be always zero, then if a mistake has occurred somewhere, the
      assembler will spot it and generate an error.  */
 
   /* If the symbol is word aligned then we might be able to reduce the
@@ -3311,7 +3373,7 @@ output_lcomm_directive (stream, name, size, rounded)
      char *name;
      int size, rounded;
 {
-  fprintf (stream, "\n\t.bss\t%c .lcomm\n", ARM_COMMENT_CHAR);
+  fprintf (stream, "\n\t.bss\t%s .lcomm\n", ASM_COMMENT_START);
   assemble_name (stream, name);
   fprintf (stream, ":\t.space\t%d\n", rounded);
   if (in_text_section ())
@@ -3461,7 +3523,7 @@ final_prescan_insn (insn, opvec, noperands)
   if (GET_CODE (insn) != JUMP_INSN)
     return;
 
-  /* This jump might be paralled with a clobber of the condition codes 
+  /* This jump might be paralleled with a clobber of the condition codes 
      the jump should always come first */
   if (GET_CODE (body) == PARALLEL && XVECLEN (body, 0) > 0)
     body = XVECEXP (body, 0, 0);
@@ -3679,7 +3741,7 @@ final_prescan_insn (insn, opvec, noperands)
 	}
       /* restore recog_operand (getting the attributes of other insns can
 	 destroy this array, but final.c assumes that it remains intact
-	 accross this call; since the insn has been recognized already we
+	 across this call; since the insn has been recognized already we
 	 call recog direct). */
       recog (PATTERN (insn), insn, NULL_PTR);
     }
