@@ -17,7 +17,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 #include <stdio.h>
 #include "config.h"
@@ -576,9 +577,15 @@ handle_pragma (file)
   if (strcmp (pbuf, "saveall") == 0)
     pragma_saveall = 1;
 
-  /* ??? This is deprecated.  Use section attributes.  */
+  /* ??? This is deprecated.  Delete for gcc 2.8.  */
   if (strcmp (pbuf, "section") == 0)
     {
+      static int printed_p = 0;
+      if (!printed_p)
+	{
+	  warning ("#pragma section is deprecated, use section attributes");
+	  printed_p = 1;
+	}
       while (c && !isalpha (c))
 	c = getc (file);
       psize = 0;
@@ -589,7 +596,7 @@ handle_pragma (file)
 	  c = getc (file);
 	}
       pbuf[psize] = 0;
-      named_section (pbuf);
+      named_section (NULL_TREE, pbuf);
     }
   ungetc (c, file);
   return c;
@@ -863,7 +870,7 @@ print_operand (file, x, code)
       break;
     case 'L':
       /* 'L' must always be used twice in a single pattern.  It generates
-	 the same lable twice, and then will generate a unique label the
+	 the same label twice, and then will generate a unique label the
 	 next time it is used.  */
       asm_fprintf (file, "tl%d", (lab++) / 2);
       break;
@@ -1106,6 +1113,7 @@ print_operand (file, x, code)
 	      fprintf (file, "%s", names_big[REGNO (x)]);
 	      break;
 	    case SImode:
+	    case SFmode:
 	      fprintf (file, "%s", names_extended[REGNO (x)]);
 	      break;
 	    default:
