@@ -1,5 +1,5 @@
 /* Subroutines used for code generation on IBM RS/6000.
-   Copyright (C) 1991, 1993, 1994 Free Software Foundation, Inc.
+   Copyright (C) 1991, 1993, 1994, 1995 Free Software Foundation, Inc.
    Contributed by Richard Kenner (kenner@vlsi1.ultra.nyu.edu)
 
 This file is part of GNU CC.
@@ -16,7 +16,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 #include <stdio.h>
 #include <ctype.h>
@@ -63,6 +64,12 @@ static int common_mode_defined;
 
 rtx rs6000_compare_op0, rs6000_compare_op1;
 int rs6000_compare_fp_p;
+
+#ifdef USING_SVR4_H
+/* Label number of label created for -mrelocatable, to call to so we can
+   get the address of the GOT section */
+int rs6000_pic_labelno;
+#endif
 
 /* Override command line options.  Mostly we process the processor
    type and sometimes adjust other TARGET_ options.  */
@@ -75,7 +82,7 @@ rs6000_override_options ()
   /* Simplify the entries below by making a mask for any POWER
      variant and any PowerPC variant.  */
 
-#define POWER_MASKS (MASK_POWER | MASK_POWER2)
+#define POWER_MASKS (MASK_POWER | MASK_POWER2 | MASK_MULTIPLE | MASK_STRING)
 #define POWERPC_MASKS (MASK_POWERPC | MASK_PPC_GPOPT \
 		       | MASK_PPC_GFXOPT | MASK_POWERPC64)
 #define POWERPC_OPT_MASKS (MASK_PPC_GPOPT | MASK_PPC_GFXOPT)
@@ -89,34 +96,43 @@ rs6000_override_options ()
     } processor_target_table[]
       = {{"common", PROCESSOR_COMMON, 0, POWER_MASKS | POWERPC_MASKS},
 	 {"power", PROCESSOR_POWER,
-	    MASK_POWER,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_MASKS | MASK_NEW_MNEMONICS},
 	 {"powerpc", PROCESSOR_POWERPC,
 	    MASK_POWERPC | MASK_NEW_MNEMONICS,
 	    POWER_MASKS | POWERPC_OPT_MASKS | MASK_POWERPC64},
 	 {"rios", PROCESSOR_RIOS1,
-	    MASK_POWER,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_MASKS | MASK_NEW_MNEMONICS},
 	 {"rios1", PROCESSOR_RIOS1,
-	    MASK_POWER,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_MASKS | MASK_NEW_MNEMONICS},
 	 {"rsc", PROCESSOR_PPC601,
-	    MASK_POWER,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_MASKS | MASK_NEW_MNEMONICS},
 	 {"rsc1", PROCESSOR_PPC601,
-	    MASK_POWER,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_MASKS | MASK_NEW_MNEMONICS},
 	 {"rios2", PROCESSOR_RIOS2,
-	    MASK_POWER | MASK_POWER2,
+	    MASK_POWER | MASK_MULTIPLE | MASK_STRING | MASK_POWER2,
 	    POWERPC_MASKS | MASK_NEW_MNEMONICS},
+	 {"403", PROCESSOR_PPC403,
+	    MASK_POWERPC | MASK_SOFT_FLOAT | MASK_NEW_MNEMONICS,
+	    POWER_MASKS | POWERPC_OPT_MASKS | MASK_POWERPC64},
+	 {"mpc403", PROCESSOR_PPC403,
+	    MASK_POWERPC | MASK_SOFT_FLOAT | MASK_NEW_MNEMONICS,
+	    POWER_MASKS | POWERPC_OPT_MASKS | MASK_POWERPC64},
+	 {"ppc403", PROCESSOR_PPC403,
+	    MASK_POWERPC | MASK_SOFT_FLOAT | MASK_NEW_MNEMONICS,
+	    POWER_MASKS | POWERPC_OPT_MASKS | MASK_POWERPC64},
 	 {"601", PROCESSOR_PPC601,
-	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS,
+	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_OPT_MASKS | MASK_POWERPC64},
 	 {"mpc601", PROCESSOR_PPC601,
-	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS,
+	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_OPT_MASKS | MASK_POWERPC64},
 	 {"ppc601", PROCESSOR_PPC601,
-	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS,
+	    MASK_POWER | MASK_POWERPC | MASK_NEW_MNEMONICS | MASK_MULTIPLE | MASK_STRING,
 	    MASK_POWER2 | POWERPC_OPT_MASKS | MASK_POWERPC64},
 	 {"603", PROCESSOR_PPC603,
 	    MASK_POWERPC | MASK_PPC_GFXOPT | MASK_NEW_MNEMONICS,
@@ -139,6 +155,9 @@ rs6000_override_options ()
 
   int ptt_size = sizeof (processor_target_table) / sizeof (struct ptt);
 
+  int multiple = TARGET_MULTIPLE;	/* save current -mmultiple/-mno-multiple status */
+  int string   = TARGET_STRING;		/* save current -mstring/-mno-string status */
+
   profile_block_flag = 0;
 
   /* Identify the processor type */
@@ -162,7 +181,54 @@ rs6000_override_options ()
 	  rs6000_cpu = PROCESSOR_DEFAULT;
 	}
     }
+
+  /* If -mmultiple or -mno-multiple was explicitly used, don't
+     override with the processor default */
+  if (TARGET_MULTIPLE_SET)
+    target_flags = (target_flags & ~MASK_MULTIPLE) | multiple;
+
+  /* If -mstring or -mno-string was explicitly used, don't
+     override with the processor default */
+  if (TARGET_STRING_SET)
+    target_flags = (target_flags & ~MASK_STRING) | string;
+
+  /* Don't allow -mmultiple or -mstring on little endian systems, because the
+     hardware doesn't support the instructions used in little endian mode */
+  if (!BYTES_BIG_ENDIAN)
+    {
+      if (TARGET_MULTIPLE)
+	{
+	  target_flags &= ~MASK_MULTIPLE;
+	  if (TARGET_MULTIPLE_SET)
+	    warning ("-mmultiple is not supported on little endian systems");
+	}
+
+      if (TARGET_STRING)
+	{
+	  target_flags &= ~MASK_STRING;
+	  if (TARGET_STRING_SET)
+	    warning ("-mstring is not supported on little endian systems");
+	}
+    }
+
+#ifdef SUBTARGET_OVERRIDE_OPTIONS
+  SUBTARGET_OVERRIDE_OPTIONS;
+#endif
 }
+
+/* Create a CONST_DOUBLE like immed_double_const, except reverse the
+   two parts of the constant if the target is little endian.  */
+
+struct rtx_def *rs6000_immed_double_const (i0, i1, mode)
+     HOST_WIDE_INT i0, i1;
+     enum machine_mode mode;
+{
+  if (! WORDS_BIG_ENDIAN)
+    return immed_double_const (i1, i0, mode);
+
+  return immed_double_const (i0, i1, mode);
+}
+
 
 /* Return non-zero if this function is known to have a null epilogue.  */
 
@@ -322,6 +388,23 @@ easy_fp_constant (op, mode)
 	  || (low != 0 && input_operand (low, word_mode)));
 }
       
+/* Return 1 if the operand is a constant whose low-order 32 bits are
+   zero.  */
+
+int
+low_32_bit_operand (op, mode)
+     register rtx op;
+     enum machine_mode mode;
+{
+  rtx low;
+
+  if (GET_CODE (op) != CONST_DOUBLE && GET_CODE (op) != CONST_INT)
+    return 0;
+
+  low = operand_subword (op, 1, 0, mode);
+  return low != 0 && GET_CODE (low) == CONST_INT && INTVAL (low) == 0;
+}
+
 /* Return 1 if the operand is either a floating-point register, a pseudo
    register, or memory.  */
 
@@ -468,6 +551,26 @@ reg_or_mem_operand (op, mode)
   return gpc_reg_operand (op, mode) || memory_operand (op, mode);
 }
 
+/* Return 1 if the operand is a general register or memory operand without
+   pre-inc or pre_dec which produces invalid form of PowerPC lwa
+   instruction.  */
+
+int
+lwa_operand (op, mode)
+     register rtx op;
+     register enum machine_mode mode;
+{
+  rtx inner = op;
+
+  if (reload_completed && GET_CODE (inner) == SUBREG)
+    inner = SUBREG_REG (inner);
+    
+  return gpc_reg_operand (inner, mode)
+    || (memory_operand (inner, mode)
+	&& GET_CODE (XEXP (inner, 0)) != PRE_INC
+	&& GET_CODE (XEXP (inner, 0)) != PRE_DEC);
+}
+
 /* Return 1 if the operand, used inside a MEM, is a valid first argument
    to CALL.  This is a SYMBOL_REF or a pseudo-register, which will be
    forced to lr.  */
@@ -533,10 +636,227 @@ input_operand (op, mode)
       && GET_CODE (op) == CONST_INT)
     return 1;
 
+  /* A SYMBOL_REF referring to the TOC is valid.  */
+  if (LEGITIMATE_CONSTANT_POOL_ADDRESS_P (op))
+    return 1;
+
   /* Otherwise, we will be doing this SET with an add, so anything valid
      for an add will be valid.  */
   return add_operand (op, mode);
 }
+
+/* Expand a block move operation, and return 1 if successful.  Return 0
+   if we should let the compiler generate normal code.
+
+   operands[0] is the destination
+   operands[1] is the source
+   operands[2] is the length
+   operands[3] is the alignment */
+
+#define MAX_MOVE_REG 4
+
+int
+expand_block_move (operands)
+     rtx operands[];
+{
+  rtx bytes_rtx	= operands[2];
+  rtx align_rtx = operands[3];
+  int constp	= (GET_CODE (bytes_rtx) == CONST_INT);
+  int align	= XINT (align_rtx, 0);
+  int bytes;
+  int offset;
+  int num_reg;
+  int i;
+  rtx src_reg;
+  rtx dest_reg;
+  rtx src_addr;
+  rtx dest_addr;
+  rtx tmp_reg;
+  rtx stores[MAX_MOVE_REG];
+  int move_bytes;
+
+  /* If this is not a fixed size move, just call memcpy */
+  if (!constp)
+    return 0;
+
+  /* Anything to move? */
+  bytes = INTVAL (bytes_rtx);
+  if (bytes <= 0)
+    return 1;
+
+  /* Don't support real large moves.  If string instructions are not used,
+     then don't generate more than 8 loads.  */
+  if (TARGET_STRING)
+    {
+      if (bytes > 4*8)
+	return 0;
+    }
+  else if (!STRICT_ALIGNMENT)
+    {
+      if (bytes > 4*8)
+	return 0;
+    }
+  else if (bytes > 8*align)
+    return 0;
+
+  /* Move the address into scratch registers.  */
+  dest_reg = copy_addr_to_reg (XEXP (operands[0], 0));
+  src_reg  = copy_addr_to_reg (XEXP (operands[1], 0));
+
+  if (TARGET_STRING)	/* string instructions are available */
+    {
+      for ( ; bytes > 0; bytes -= move_bytes)
+	{
+#if 0
+	  /* XXX Don't move so many bytes right now, it causes the compiler to not bootstrap */
+	  if (bytes > 24		/* move up to 32 bytes at a time */
+	      && !fixed_regs[5]
+	      && !fixed_regs[6]
+	      && !fixed_regs[7]
+	      && !fixed_regs[8]
+	      && !fixed_regs[9]
+	      && !fixed_regs[10]
+	      && !fixed_regs[11]
+	      && !fixed_regs[12])
+	    {
+	      move_bytes = (bytes > 32) ? 32 : bytes;
+	      emit_insn (gen_movstrsi_8reg (dest_reg,
+					    src_reg,
+					    GEN_INT ((move_bytes == 32) ? 0 : move_bytes),
+					    align_rtx));
+	    }
+	  else if (bytes > 16	/* move up to 24 bytes at a time */
+		   && !fixed_regs[7]
+		   && !fixed_regs[8]
+		   && !fixed_regs[9]
+		   && !fixed_regs[10]
+		   && !fixed_regs[11]
+		   && !fixed_regs[12])
+	    {
+	      move_bytes = (bytes > 24) ? 24 : bytes;
+	      emit_insn (gen_movstrsi_6reg (dest_reg,
+					    src_reg,
+					    GEN_INT (move_bytes),
+					    align_rtx));
+	    }
+	  else if (bytes > 8	/* move up to 16 bytes at a time */
+		   && !fixed_regs[9]
+		   && !fixed_regs[10]
+		   && !fixed_regs[11]
+		   && !fixed_regs[12])
+	    {
+	      move_bytes = (bytes > 16) ? 16 : bytes;
+	      emit_insn (gen_movstrsi_4reg (dest_reg,
+					    src_reg,
+					    GEN_INT (move_bytes),
+					    align_rtx));
+	    }
+	  else
+#endif
+	    if (bytes > 4 && !TARGET_64BIT)
+	    {			/* move up to 8 bytes at a time */
+	      move_bytes = (bytes > 8) ? 8 : bytes;
+	      emit_insn (gen_movstrsi_2reg (dest_reg,
+					    src_reg,
+					    GEN_INT (move_bytes),
+					    align_rtx));
+	    }
+	  else if (bytes >= 4 && (align >= 4 || !STRICT_ALIGNMENT))
+	    {			/* move 4 bytes */
+	      move_bytes = 4;
+	      tmp_reg = gen_reg_rtx (SImode);
+	      emit_move_insn (tmp_reg, gen_rtx (MEM, SImode, src_reg));
+	      emit_move_insn (gen_rtx (MEM, SImode, dest_reg), tmp_reg);
+	    }
+	  else if (bytes == 2 && (align >= 2 || !STRICT_ALIGNMENT))
+	    {			/* move 2 bytes */
+	      move_bytes = 2;
+	      tmp_reg = gen_reg_rtx (HImode);
+	      emit_move_insn (tmp_reg, gen_rtx (MEM, HImode, src_reg));
+	      emit_move_insn (gen_rtx (MEM, HImode, dest_reg), tmp_reg);
+	    }
+	  else if (bytes == 1)	/* move 1 byte */
+	    {
+	      move_bytes = 1;
+	      tmp_reg = gen_reg_rtx (QImode);
+	      emit_move_insn (tmp_reg, gen_rtx (MEM, QImode, src_reg));
+	      emit_move_insn (gen_rtx (MEM, QImode, dest_reg), tmp_reg);
+	    }
+	  else
+	    {			/* move up to 4 bytes at a time */
+	      move_bytes = (bytes > 4) ? 4 : bytes;
+	      emit_insn (gen_movstrsi_1reg (dest_reg,
+					    src_reg,
+					    GEN_INT (move_bytes),
+					    align_rtx));
+	    }
+
+	  if (bytes > move_bytes)
+	    {
+	      emit_insn (gen_addsi3 (src_reg, src_reg, GEN_INT (move_bytes)));
+	      emit_insn (gen_addsi3 (dest_reg, dest_reg, GEN_INT (move_bytes)));
+	    }
+	}
+    }
+
+  else			/* string instructions not available */
+    {
+      num_reg = offset = 0;
+      for ( ; bytes > 0; (bytes -= move_bytes), (offset += move_bytes))
+	{
+	  /* Calculate the correct offset for src/dest */
+	  if (offset == 0)
+	    {
+	      src_addr  = src_reg;
+	      dest_addr = dest_reg;
+	    }
+	  else
+	    {
+	      src_addr  = gen_rtx (PLUS, Pmode, src_reg,  GEN_INT (offset));
+	      dest_addr = gen_rtx (PLUS, Pmode, dest_reg, GEN_INT (offset));
+	    }
+
+	  /* Generate the appropriate load and store, saving the stores for later */
+	  if (bytes >= 4 && (align >= 4 || !STRICT_ALIGNMENT))
+	    {
+	      move_bytes = 4;
+	      tmp_reg = gen_reg_rtx (SImode);
+	      emit_insn (gen_movsi (tmp_reg, gen_rtx (MEM, SImode, src_addr)));
+	      stores[ num_reg++ ] = gen_movsi (gen_rtx (MEM, SImode, dest_addr), tmp_reg);
+	    }
+	  else if (bytes >= 2 && (align >= 2 || !STRICT_ALIGNMENT))
+	    {
+	      move_bytes = 2;
+	      tmp_reg = gen_reg_rtx (HImode);
+	      emit_insn (gen_movhi (tmp_reg, gen_rtx (MEM, HImode, src_addr)));
+	      stores[ num_reg++ ] = gen_movhi (gen_rtx (MEM, HImode, dest_addr), tmp_reg);
+	    }
+	  else
+	    {
+	      move_bytes = 1;
+	      tmp_reg = gen_reg_rtx (QImode);
+	      emit_insn (gen_movqi (tmp_reg, gen_rtx (MEM, QImode, src_addr)));
+	      stores[ num_reg++ ] = gen_movqi (gen_rtx (MEM, QImode, dest_addr), tmp_reg);
+	    }
+
+	  if (num_reg >= MAX_MOVE_REG)
+	    {
+	      for (i = 0; i < num_reg; i++)
+		emit_insn (stores[i]);
+	      num_reg = 0;
+	    }
+	}
+
+      if (num_reg > 0)
+	{
+	  for (i = 0; i < num_reg; i++)
+	    emit_insn (stores[i]);
+	}
+    }
+
+  return 1;
+}
+
 
 /* Return 1 if OP is a load multiple operation.  It is known to be a
    PARALLEL and the first section will be tested.  */
@@ -719,6 +1039,78 @@ includes_rshift_p (shiftop, andop)
 
   return (INTVAL (andop) & ~ shift_mask) == 0;
 }
+
+/* Return 1 if REGNO (reg1) == REGNO (reg2) - 1 making them candidates
+   for lfq and stfq insns.
+
+   Note reg1 and reg2 *must* be hard registers.  To be sure we will
+   abort if we are passed pseudo registers.  */
+
+int
+registers_ok_for_quad_peep (reg1, reg2)
+     rtx reg1, reg2;
+{
+  /* We might have been passed a SUBREG.  */
+  if (GET_CODE (reg1) != REG || GET_CODE (reg2) != REG) 
+    return 0;
+
+  return (REGNO (reg1) == REGNO (reg2) - 1);
+}
+
+/* Return 1 if addr1 and addr2 are suitable for lfq or stfq insn.  addr1 and
+   addr2 must be in consecutive memory locations (addr2 == addr1 + 8).  */
+
+int
+addrs_ok_for_quad_peep (addr1, addr2)
+     register rtx addr1;
+     register rtx addr2;
+{
+  int reg1;
+  int offset1;
+
+  /* Extract an offset (if used) from the first addr.  */
+  if (GET_CODE (addr1) == PLUS)
+    {
+      /* If not a REG, return zero.  */
+      if (GET_CODE (XEXP (addr1, 0)) != REG)
+	return 0;
+      else
+	{
+          reg1 = REGNO (XEXP (addr1, 0));
+	  /* The offset must be constant!  */
+	  if (GET_CODE (XEXP (addr1, 1)) != CONST_INT)
+            return 0;
+          offset1 = INTVAL (XEXP (addr1, 1));
+	}
+    }
+  else if (GET_CODE (addr1) != REG)
+    return 0;
+  else
+    {
+      reg1 = REGNO (addr1);
+      /* This was a simple (mem (reg)) expression.  Offset is 0.  */
+      offset1 = 0;
+    }
+
+  /* Make sure the second address is a (mem (plus (reg) (const_int).  */
+  if (GET_CODE (addr2) != PLUS)
+    return 0;
+
+  if (GET_CODE (XEXP (addr2, 0)) != REG
+      || GET_CODE (XEXP (addr2, 1)) != CONST_INT)
+    return 0;
+
+  if (reg1 != REGNO (XEXP (addr2, 0)))
+    return 0;
+
+  /* The offset for the second addr must be 8 more than the first addr.  */
+  if (INTVAL (XEXP (addr2, 1)) != offset1 + 8)
+    return 0;
+
+  /* All the tests passed.  addr1 and addr2 are valid for lfq or stfq
+     instructions.  */
+  return 1;
+}
 
 /* Return the register class of a scratch register needed to copy IN into
    or out of a register in CLASS in MODE.  If it can be done directly,
@@ -838,6 +1230,11 @@ print_operand (file, x, code)
       asm_fprintf (file, RS6000_CALL_GLUE);
       return;
 
+    case '*':
+      /* Write the register number of the TOC register.  */
+      fputs (TARGET_MINIMAL_TOC ? "30" : "2", file);
+      return;
+
     case 'A':
       /* If X is a constant integer whose low-order 5 bits are zero,
 	 write 'l'.  Otherwise, write 'r'.  This is a kludge to fix a bug
@@ -939,14 +1336,6 @@ print_operand (file, x, code)
 	print_operand (file, x, 0);
       return;
 
-    case 'H':
-      /* X must be a constant.  Output the low order 5 bits plus 24.  */
-      if (! INT_P (x))
-	output_operand_lossage ("invalid %%H value");
-
-      fprintf (file, "%d", (INT_LOWPART (x) + 24) & 31);
-      return;
-
     case 'I':
       /* Print `i' if this is a constant, else nothing.  */
       if (INT_P (x))
@@ -1121,14 +1510,6 @@ print_operand (file, x, code)
       fprintf (file, "%d", (32 - INT_LOWPART (x)) & 31);
       return;
 
-    case 'S':
-      /* Low 5 bits of 31 - value */
-      if (! INT_P (x))
-	output_operand_lossage ("invalid %%S value");
-
-      fprintf (file, "%d", (31 - INT_LOWPART (x)) & 31);
-      return;
-
     case 't':
       /* Write 12 if this jump operation will branch if true, 4 otherwise. 
 	 All floating-point operations except NE branch true and integer
@@ -1220,11 +1601,15 @@ print_operand (file, x, code)
     case 'z':
       /* X is a SYMBOL_REF.  Write out the name preceded by a
 	 period and without any trailing data in brackets.  Used for function
-	 names.  */
+	 names.  If we are configured for System V (or the embedded ABI) on
+	 the PowerPC, do not emit the period, since those systems do not use
+	 TOCs and the like.  */
       if (GET_CODE (x) != SYMBOL_REF)
 	abort ();
 
+#ifndef USING_SVR4_H
       putc ('.', file);
+#endif
       RS6000_OUTPUT_BASENAME (file, XSTR (x, 0));
       return;
 
@@ -1402,6 +1787,74 @@ rs6000_pushes_stack ()
 	  || rs6000_makes_calls ());
 }
 
+#ifdef USING_SVR4_H
+/* Write out a System V.4 style traceback table before the prologue
+
+   At present, only emit the basic tag table (ie, do not emit tag_types other
+   than 0, which might use more than 1 tag word).
+
+   The first tag word looks like:
+
+    0			1		    2			3
+    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+   |         0 |ver| tag |e|s| alloca  | # fprs  | # gprs  |s|l|c|f|
+   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+
+*/
+
+void
+svr4_traceback (file, name, decl)
+     FILE *file;
+     tree name, decl;
+{
+
+  int first_reg		= first_reg_to_save ();
+  int first_fp_reg	= first_fp_reg_to_save ();
+  int pushes_stack	= rs6000_pushes_stack ();
+  long tag;
+  long version		= 0;			/* version number */
+  long tag_type		= 0;			/* function type */
+  long extended_tag	= 0;			/* additional tag words needed */
+  long spare		= 0;			/* reserved for future use */
+  long alloca_reg;				/* stack/frame register */
+  long fpr_max		= 64 - first_fp_reg;	/* # of floating point registers saved */
+  long gpr_max		= 32 - first_reg;	/* # of general purpose registers saved */
+  long sp_max;					/* 1 if the function acquires a stack frame */
+  long lr_max;					/* 1 if the function stores the link register */
+  long cr_max;					/* 1 if the function has a CR save word */
+  long fpscr_max	= 0;			/* 1 if the function has a FPSCR save word */
+
+  if (frame_pointer_needed)
+    alloca_reg = 31;
+
+  else if (pushes_stack != 0)
+    alloca_reg = 1;
+
+  else
+    alloca_reg = 0;
+
+  lr_max = (regs_ever_live[65] || first_fp_reg < 62 || profile_flag);
+  cr_max = (must_save_cr () != 0);
+  sp_max = (pushes_stack != 0);
+
+  tag = (((version & 3) << 24)
+	 | ((tag_type & 7) << 21)
+	 | ((extended_tag & 1) << 20)
+	 | ((spare & 1) << 19)
+	 | ((alloca_reg & 0x1f) << 14)
+	 | ((fpr_max & 0x1f) << 9)
+	 | ((gpr_max & 0x1f) << 4)
+	 | ((sp_max & 1) << 3)
+	 | ((lr_max & 1) << 2)
+	 | ((cr_max & 1) << 1)
+	 | ((fpscr_max & 1) << 0));
+	   
+  fprintf (file, "\t.long 0x%lx\n", tag);
+}
+
+#endif /* USING_SVR4_H */
+
 /* Write function prologue.  */
 
 void
@@ -1414,15 +1867,19 @@ output_prolog (file, size)
   int first_fp_reg = first_fp_reg_to_save ();
   int basic_size = rs6000_sa_size ();
   int total_size = (basic_size + size + current_function_outgoing_args_size);
+  char buf[256];
 
   /* Round size to multiple of 8 bytes.  */
   total_size = (total_size + 7) & ~7;
 
   /* Write .extern for any function we will call to save and restore fp
      values.  */
+#ifndef USING_SVR4_H
   if (first_fp_reg < 62)
-    fprintf (file, "\t.extern ._savef%d\n\t.extern ._restf%d\n",
-	     first_fp_reg - 32, first_fp_reg - 32);
+    fprintf (file, "\t.extern %s%d%s\n\t.extern %s%d%s\n",
+	     SAVE_FP_PREFIX, first_fp_reg - 32, SAVE_FP_SUFFIX,
+	     RESTORE_FP_PREFIX, first_fp_reg - 32, RESTORE_FP_SUFFIX);
+#endif
 
   /* Write .extern for truncation routines, if needed.  */
   if (rs6000_trunc_used && ! trunc_defined)
@@ -1443,11 +1900,22 @@ output_prolog (file, size)
       common_mode_defined = 1;
     }
 
+#ifdef USING_SVR4_H
+  /* If we have a relocatable GOT section, we need to save the LR. */
+  if (TARGET_RELOCATABLE && get_pool_size () != 0)
+    regs_ever_live[65] = 1;
+#endif
+
   /* If we have to call a function to save fpr's, or if we are doing profiling,
      then we will be using LR.  */
-  if (first_fp_reg < 62 || profile_flag)
+  if (profile_flag)
     regs_ever_live[65] = 1;
 
+#ifndef USING_SVR4_H
+  if (first_fp_reg < 62)
+    regs_ever_live[65] = 1;
+#endif
+
   /* If we use the link register, get it into r0.  */
   if (regs_ever_live[65])
     asm_fprintf (file, "\tmflr 0\n");
@@ -1464,10 +1932,22 @@ output_prolog (file, size)
   else if (first_fp_reg == 63)
     asm_fprintf (file, "\tstfd 31,-8(1)\n");
   else if (first_fp_reg != 64)
-    asm_fprintf (file, "\tbl ._savef%d\n", first_fp_reg - 32);
+    {
+#ifndef USING_SVR4_H
+      asm_fprintf (file, "\tbl %s%d%s\n", SAVE_FP_PREFIX, first_fp_reg - 32, SAVE_FP_SUFFIX);
+#else
+      int regno, loc;
+
+      for (regno = first_fp_reg,
+	   loc = - (64 - first_fp_reg) * 8;
+	   regno < 64;
+	   regno++, loc += 8)
+	asm_fprintf (file, "\tstfd %d,%d(1)\n", regno - 32, loc);
+#endif
+    }
 
   /* Now save gpr's.  */
-  if (! TARGET_POWER || first_reg == 31)
+  if (! TARGET_MULTIPLE || first_reg == 31)
     {
       int regno, loc;
 
@@ -1497,7 +1977,7 @@ output_prolog (file, size)
 	asm_fprintf (file, "\t{stu|stwu} 1,%d(1)\n", - total_size);
       else
 	{
-	  asm_fprintf (file, "\t{cau 0,0,%d|lis 0,%d}\n\t{oril|ori} 0,0,%d\n",
+	  asm_fprintf (file, "\t{liu|lis} 0,%d\n\t{oril|ori} 0,0,%d\n",
 		   (total_size >> 16) & 0xffff, total_size & 0xffff);
 	  if (TARGET_POWERPC)
 	    asm_fprintf (file, "\tsubf 12,0,1\n");
@@ -1515,12 +1995,54 @@ output_prolog (file, size)
      TOC_TABLE address into register 30.  */
   if (TARGET_MINIMAL_TOC && get_pool_size () != 0)
     {
-      char buf[100];
+      char buf[256];
 
-      ASM_GENERATE_INTERNAL_LABEL (buf, "LCTOC", 0);
-      asm_fprintf (file, "\t{l|lwz} 30,");
-      assemble_name (file, buf);
-      asm_fprintf (file, "(2)\n");
+#ifdef USING_SVR4_H
+      if (TARGET_RELOCATABLE)
+	{
+	  ASM_GENERATE_INTERNAL_LABEL (buf, "LCF", rs6000_pic_labelno);
+	  fprintf (file, "\tbl ");
+	  assemble_name (file, buf);
+	  fprintf (file, "\n");
+
+	  ASM_OUTPUT_INTERNAL_LABEL (file, "LCF", rs6000_pic_labelno);
+	  fprintf (file, "\tmflr 30\n");
+
+	  if (TARGET_POWERPC64)
+	    fprintf (file, "\tld 0,");
+	  else if (TARGET_NEW_MNEMONICS)
+	    fprintf (file, "\tlwz 0,");
+	  else
+	    fprintf (file, "\tl 0,");
+
+	  fprintf (file, "(");
+	  ASM_GENERATE_INTERNAL_LABEL (buf, "LCL", rs6000_pic_labelno);
+	  assemble_name (file, buf);
+	  fprintf (file, "-");
+	  ASM_GENERATE_INTERNAL_LABEL (buf, "LCF", rs6000_pic_labelno);
+	  assemble_name (file, buf);
+	  fprintf (file, ")(30)\n");
+	  asm_fprintf (file, "\t{cax|add} 30,0,30\n");
+	  rs6000_pic_labelno++;
+	}
+      else if (TARGET_NO_TOC)
+	{
+	  ASM_GENERATE_INTERNAL_LABEL (buf, "LCTOC", 1);
+	  asm_fprintf (file, "\t{cau|addis} 30,0,");
+	  assemble_name (file, buf);
+	  asm_fprintf (file, "@ha\n");
+	  asm_fprintf (file, "\t{cal|addi} 30,30,");
+	  assemble_name (file, buf);
+	  asm_fprintf (file, "@l\n");
+	}
+      else
+#endif /* USING_SVR4_H */
+	{
+	  ASM_GENERATE_INTERNAL_LABEL (buf, "LCTOC", 0);
+	  asm_fprintf (file, "\t{l|lwz} 30,");
+	  assemble_name (file, buf);
+	  asm_fprintf (file, "(2)\n");
+	}
     }
 }
 
@@ -1569,7 +2091,7 @@ output_epilog (file, size)
 	asm_fprintf (file, "\tmtlr 0\n");
 
       /* Restore gpr's.  */
-      if (! TARGET_POWER || first_reg == 31)
+      if (! TARGET_MULTIPLE || first_reg == 31)
 	{
 	  int regno, loc;
 
@@ -1601,7 +2123,21 @@ output_epilog (file, size)
       /* If we have to restore more than two FP registers, branch to the
 	 restore function.  It will return to our caller.  */
       if (first_fp_reg < 62)
-	asm_fprintf (file, "\tb ._restf%d\n", first_fp_reg - 32);
+	{
+#ifndef USING_SVR4_H
+	  asm_fprintf (file, "\tb %s%d%s\n", RESTORE_FP_PREFIX, first_fp_reg - 32, RESTORE_FP_SUFFIX);
+#else
+	  int regno, loc;
+
+	  for (regno = first_fp_reg,
+	       loc = - (64 - first_fp_reg) * 8;
+	       regno < 64;
+	       regno++, loc += 8)
+	    asm_fprintf (file, "\tlfd %d,%d(1)\n", regno - 32, loc);
+
+	  asm_fprintf (file, "\t{br|blr}\n");
+#endif
+	}
       else
 	asm_fprintf (file, "\t{br|blr}\n");
     }
@@ -1616,7 +2152,11 @@ output_epilog (file, size)
      middle, and the two halves are placed at locations far apart in
      memory.''  The traceback table has this property, since it
      includes the offset from the start of the function to the
-     traceback table itself.  */
+     traceback table itself.
+
+     System V.4 Powerpc's (and the embedded ABI derived from it) use a
+     different traceback table located before the prologue.  */
+#ifndef USING_SVR4_H
   if (! flag_inhibit_size_directive)
     {
       char *fname = XSTR (XEXP (DECL_RTL (current_function_decl), 0), 0);
@@ -1794,6 +2334,7 @@ output_epilog (file, size)
       if (frame_pointer_needed)
 	fprintf (file, "\t.byte 31\n");
     }
+#endif /* !USING_SVR4_H */
 }
 
 /* Output a TOC entry.  We derive the entry name from what is
@@ -1810,24 +2351,35 @@ output_toc (file, x, labelno)
   rtx base = x;
   int offset = 0;
 
-  ASM_OUTPUT_INTERNAL_LABEL (file, "LC", labelno);
+#ifdef USING_SVR4_H
+  if (TARGET_MINIMAL_TOC)
+    {
+      ASM_OUTPUT_INTERNAL_LABEL_PREFIX (file, "LC");
+      fprintf (file, "%d = .-", labelno);
+      ASM_OUTPUT_INTERNAL_LABEL_PREFIX (file, "LCTOC");
+      fprintf (file, "1\n");
+    }
+  else
+#endif /* USING_SVR4_H */
+    ASM_OUTPUT_INTERNAL_LABEL (file, "LC", labelno);
 
   /* Handle FP constants specially.  Note that if we have a minimal
      TOC, things we put here aren't actually in the TOC, so we can allow
      FP constants.  */
   if (GET_CODE (x) == CONST_DOUBLE
       && GET_MODE (x) == DFmode
-      && TARGET_FLOAT_FORMAT == HOST_FLOAT_FORMAT
-      && BITS_PER_WORD == HOST_BITS_PER_INT
       && ! (TARGET_NO_FP_IN_TOC && ! TARGET_MINIMAL_TOC))
     {
+      REAL_VALUE_TYPE r;
+      long l[2];
+
+      REAL_VALUE_FROM_CONST_DOUBLE (r, x);
+      REAL_VALUE_TO_TARGET_DOUBLE (r, l);
       if (TARGET_MINIMAL_TOC)
-	fprintf (file, "\t.long %d\n\t.long %d\n",
-		 CONST_DOUBLE_LOW (x), CONST_DOUBLE_HIGH (x));
+	fprintf (file, "\t.long %ld\n\t.long %ld\n", l[0], l[1]);
       else
-	fprintf (file, "\t.tc FD_%x_%x[TC],%d,%d\n",
-		 CONST_DOUBLE_LOW (x), CONST_DOUBLE_HIGH (x),
-		 CONST_DOUBLE_LOW (x), CONST_DOUBLE_HIGH (x));
+	fprintf (file, "\t.tc FD_%lx_%lx[TC],%ld,%ld\n",
+		 l[0], l[1], l[0], l[1]);
       return;
     }
   else if (GET_CODE (x) == CONST_DOUBLE && GET_MODE (x) == SFmode
@@ -2010,6 +2562,9 @@ output_function_profiler (file, labelno)
   FILE *file;
   int labelno;
 {
+#ifdef USING_SVR4_H
+  abort ();
+#else
   /* The last used parameter register.  */
   int last_parm_reg;
   int i, j;
@@ -2061,6 +2616,7 @@ output_function_profiler (file, labelno)
 
   for (i = 3, j = 30; i <= last_parm_reg; i++, j--)
     fprintf (file, "\tai %d,%d,0\n", i, j);
+#endif
 }
 
 /* Adjust the cost of a scheduling dependency.  Return the new cost of
