@@ -1,5 +1,5 @@
 /* Definitions of target machine for GNU compiler, for IBM RS/6000.
-   Copyright (C) 1992, 1993, 1994 Free Software Foundation, Inc.
+   Copyright (C) 1992, 1993, 1994, 1995 Free Software Foundation, Inc.
    Contributed by Richard Kenner (kenner@vlsi1.ultra.nyu.edu)
 
 This file is part of GNU CC.
@@ -16,7 +16,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 
 /* Note that some other tm.h files include this one and then override
@@ -45,6 +46,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
    target switches.  */
 
 #define CPP_SPEC "\
+%{posix: -D_POSIX_SOURCE} \
 %{!mcpu*: \
   %{mpower: %{!mpower2: -D_ARCH_PWR}} \
   %{mpower2: -D_ARCH_PWR2} \
@@ -59,6 +61,9 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 %{mcpu=rios2: -D_ARCH_PWR2} \
 %{mcpu=rsc: -D_ARCH_PWR} \
 %{mcpu=rsc1: -D_ARCH_PWR} \
+%{mcpu=403: -D_ARCH_PPC} \
+%{mcpu=mpc403: -D_ARCH_PPC} \
+%{mcpu=ppc403: -D_ARCH_PPC} \
 %{mcpu=601: -D_ARCH_PPC -D_ARCH_PWR} \
 %{mcpu=mpc601: -D_ARCH_PPC -D_ARCH_PWR} \
 %{mcpu=ppc601: -D_ARCH_PPC -D_ARCH_PWR} \
@@ -77,17 +82,22 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
    used by the AIX cc command.  So we use it here.
 
    -bnodelcsect undoes a poor choice of default relating to multiply-defined
-   csects.  See AIX documentation for more information about this.  */
+   csects.  See AIX documentation for more information about this.
+
+   -bM:SRE tells the linker that the output file is Shared REusable.  Note
+   that to actually build a shared library you will also need to specify an
+   export list with the -Wl,-bE option.  */
 
 #define LINK_SPEC "-T512 -H512 %{!r:-btextro} -bhalt:4 -bnodelcsect\
-   %{static:-bnso -bI:/lib/syscalls.exp} %{g*:-bexport:/usr/lib/libg.exp}"
+   %{static:-bnso -bI:/lib/syscalls.exp} \
+   %{!shared:%{g*:-bexport:/usr/lib/libg.exp}} %{shared:-bM:SRE}"
 
 /* Profiled library versions are used by linking with special directories.  */
 #define LIB_SPEC "%{pg:-L/lib/profiled -L/usr/lib/profiled}\
-   %{p:-L/lib/profiled -L/usr/lib/profiled} %{g*:-lg} -lc"
+   %{p:-L/lib/profiled -L/usr/lib/profiled} %{!shared:%{g*:-lg}} -lc"
 
 /* gcc must do the search itself to find libgcc.a, not use -l.  */
-#define LINK_LIBGCC_SPECIAL_1
+#define LIBGCC_SPEC "libgcc.a%s"
 
 /* Don't turn -B into -L if the argument specifies a relative file name.  */
 #define RELATIVE_PREFIX_NOT_LINKDIR
@@ -134,16 +144,38 @@ extern int target_flags;
    function, and one less allocatable register.  */
 #define MASK_MINIMAL_TOC	0x200
 
-#define TARGET_POWER			(target_flags & MASK_POWER)
-#define TARGET_POWER2			(target_flags & MASK_POWER2)
-#define TARGET_POWERPC			(target_flags & MASK_POWERPC)
-#define TARGET_PPC_GPOPT		(target_flags & MASK_PPC_GPOPT)
-#define TARGET_PPC_GFXOPT		(target_flags & MASK_PPC_GFXOPT)
-#define TARGET_POWERPC64		(target_flags & MASK_POWERPC64)
-#define TARGET_NEW_MNEMONICS		(target_flags & MASK_NEW_MNEMONICS)
-#define TARGET_NO_FP_IN_TOC		(target_flags & MASK_NO_FP_IN_TOC)
-#define TARGET_NO_SUM_IN_TOC		(target_flags & MASK_NO_SUM_IN_TOC)
-#define TARGET_MINIMAL_TOC		(target_flags & MASK_MINIMAL_TOC)
+/* Nonzero for the 64bit model: ints, longs, and pointers are 64 bits.  */
+#define MASK_64BIT		0x400
+
+/* Disable use of FPRs.  */
+#define MASK_SOFT_FLOAT		0x800
+
+/* Enable load/store multiple, even on powerpc */
+#define	MASK_MULTIPLE		0x1000
+#define	MASK_MULTIPLE_SET	0x2000
+
+/* Use string instructions for block moves */
+#define MASK_STRING		0x4000
+#define MASK_STRING_SET		0x8000
+
+#define TARGET_POWER		(target_flags & MASK_POWER)
+#define TARGET_POWER2		(target_flags & MASK_POWER2)
+#define TARGET_POWERPC		(target_flags & MASK_POWERPC)
+#define TARGET_PPC_GPOPT	(target_flags & MASK_PPC_GPOPT)
+#define TARGET_PPC_GFXOPT	(target_flags & MASK_PPC_GFXOPT)
+#define TARGET_POWERPC64	(target_flags & MASK_POWERPC64)
+#define TARGET_NEW_MNEMONICS	(target_flags & MASK_NEW_MNEMONICS)
+#define TARGET_NO_FP_IN_TOC	(target_flags & MASK_NO_FP_IN_TOC)
+#define TARGET_NO_SUM_IN_TOC	(target_flags & MASK_NO_SUM_IN_TOC)
+#define TARGET_MINIMAL_TOC	(target_flags & MASK_MINIMAL_TOC)
+#define TARGET_64BIT		(target_flags & MASK_64BIT)
+#define TARGET_SOFT_FLOAT	(target_flags & MASK_SOFT_FLOAT)
+#define	TARGET_MULTIPLE		(target_flags & MASK_MULTIPLE)
+#define	TARGET_MULTIPLE_SET	(target_flags & MASK_MULTIPLE_SET)
+#define TARGET_STRING		(target_flags & MASK_STRING)
+#define TARGET_STRING_SET	(target_flags & MASK_STRING_SET)
+
+#define TARGET_HARD_FLOAT	(! TARGET_SOFT_FLOAT)
 
 /* Run-time compilation parameters selecting different hardware subsets.
 
@@ -153,37 +185,54 @@ extern int target_flags;
    where VALUE is the bits to set or minus the bits to clear.
    An empty string NAME is used to identify the default VALUE.  */
 
-#define TARGET_SWITCHES						\
- {{"power",		MASK_POWER},				\
-  {"power2",		MASK_POWER | MASK_POWER2},		\
-  {"no-power2",		- MASK_POWER2},				\
-  {"no-power",		- (MASK_POWER | MASK_POWER2)},		\
-  {"powerpc",		MASK_POWERPC},				\
-  {"no-powerpc",	- (MASK_POWERPC | MASK_PPC_GPOPT 	\
-			   | MASK_PPC_GFXOPT | MASK_POWERPC64)}, \
-  {"powerpc-gpopt",	MASK_POWERPC | MASK_PPC_GPOPT},		\
-  {"no-powerpc-gpopt",	- MASK_PPC_GPOPT},			\
-  {"powerpc-gfxopt",	MASK_POWERPC | MASK_PPC_GFXOPT},	\
-  {"no-powerpc-gfxopt",	- MASK_PPC_GFXOPT},			\
-  {"new-mnemonics",	MASK_NEW_MNEMONICS},			\
-  {"old-mnemonics",	-MASK_NEW_MNEMONICS},			\
-  {"full-toc",		- (MASK_NO_FP_IN_TOC | MASK_NO_SUM_IN_TOC \
-			   | MASK_MINIMAL_TOC)},		\
-  {"fp-in-toc",		- MASK_NO_FP_IN_TOC},			\
-  {"no-fp-in-toc",	MASK_NO_FP_IN_TOC},			\
-  {"sum-in-toc",	- MASK_NO_SUM_IN_TOC},			\
-  {"no-sum-in-toc",	MASK_NO_SUM_IN_TOC},			\
-  {"minimal-toc",	MASK_MINIMAL_TOC},			\
-  {"minimal-toc",	- (MASK_NO_FP_IN_TOC | MASK_NO_SUM_IN_TOC)}, \
-  {"no-minimal-toc",	- MASK_MINIMAL_TOC},			\
+/* This is meant to be redefined in the host dependent files */
+#ifndef SUBTARGET_SWITCHES
+#define SUBTARGET_SWITCHES
+#endif
+
+#define TARGET_SWITCHES							\
+ {{"power",		MASK_POWER  | MASK_MULTIPLE | MASK_STRING},	\
+  {"power2",		(MASK_POWER | MASK_MULTIPLE | MASK_STRING	\
+			 | MASK_POWER2)},				\
+  {"no-power2",		- MASK_POWER2},					\
+  {"no-power",		- (MASK_POWER | MASK_POWER2 | MASK_MULTIPLE	\
+			   | MASK_STRING)},				\
+  {"powerpc",		MASK_POWERPC},					\
+  {"no-powerpc",	- (MASK_POWERPC | MASK_PPC_GPOPT		\
+			   | MASK_PPC_GFXOPT | MASK_POWERPC64)},	\
+  {"powerpc-gpopt",	MASK_POWERPC | MASK_PPC_GPOPT},			\
+  {"no-powerpc-gpopt",	- MASK_PPC_GPOPT},				\
+  {"powerpc-gfxopt",	MASK_POWERPC | MASK_PPC_GFXOPT},		\
+  {"no-powerpc-gfxopt",	- MASK_PPC_GFXOPT},				\
+  {"new-mnemonics",	MASK_NEW_MNEMONICS},				\
+  {"old-mnemonics",	-MASK_NEW_MNEMONICS},				\
+  {"full-toc",		- (MASK_NO_FP_IN_TOC | MASK_NO_SUM_IN_TOC	\
+			   | MASK_MINIMAL_TOC)},			\
+  {"fp-in-toc",		- MASK_NO_FP_IN_TOC},				\
+  {"no-fp-in-toc",	MASK_NO_FP_IN_TOC},				\
+  {"sum-in-toc",	- MASK_NO_SUM_IN_TOC},				\
+  {"no-sum-in-toc",	MASK_NO_SUM_IN_TOC},				\
+  {"minimal-toc",	MASK_MINIMAL_TOC},				\
+  {"minimal-toc",	- (MASK_NO_FP_IN_TOC | MASK_NO_SUM_IN_TOC)},	\
+  {"no-minimal-toc",	- MASK_MINIMAL_TOC},				\
+  {"hard-float",	- MASK_SOFT_FLOAT},				\
+  {"soft-float",	MASK_SOFT_FLOAT},				\
+  {"multiple",		MASK_MULTIPLE | MASK_MULTIPLE_SET},		\
+  {"no-multiple",	- MASK_MULTIPLE},				\
+  {"no-multiple",	MASK_MULTIPLE_SET},				\
+  {"string",		MASK_STRING | MASK_STRING_SET},			\
+  {"no-string",		- MASK_STRING},					\
+  {"no-string",		MASK_STRING_SET},				\
+  SUBTARGET_SWITCHES							\
   {"",			TARGET_DEFAULT}}
 
-#define TARGET_DEFAULT MASK_POWER
+#define TARGET_DEFAULT (MASK_POWER | MASK_MULTIPLE | MASK_STRING)
 
 /* Processor type.  */
 enum processor_type
  {PROCESSOR_RIOS1,
   PROCESSOR_RIOS2,
+  PROCESSOR_PPC403,
   PROCESSOR_PPC601,
   PROCESSOR_PPC603,
   PROCESSOR_PPC604,
@@ -243,6 +292,9 @@ extern char *rs6000_cpu_string;
 
 /* target machine storage layout */
 
+/* Define to support cross compilation to an RS6000 target.  */
+#define REAL_ARITHMETIC
+
 /* Define this macro if it is advisable to hold scalars in registers
    in a wider mode than that declared by the program.  In such cases, 
    the value is constrained to be within the bounds of the declared
@@ -277,10 +329,13 @@ extern char *rs6000_cpu_string;
    Note that this is not necessarily the width of data type `int';
    if using 16-bit ints on a 68000, this would still be 32.
    But on a machine with 16-bit registers, this would be 16.  */
-#define BITS_PER_WORD 32
+#define BITS_PER_WORD (TARGET_POWERPC64 ? 64 : 32)
+#define MAX_BITS_PER_WORD 64
 
 /* Width of a word, in units (bytes).  */
-#define UNITS_PER_WORD 4
+#define UNITS_PER_WORD (TARGET_POWERPC64 ? 8 : 4)
+#define MIN_UNITS_PER_WORD 4
+#define UNITS_PER_FP_WORD 8
 
 /* Type used for ptrdiff_t, as a string used in a declaration.  */
 #define PTRDIFF_TYPE "int"
@@ -291,12 +346,56 @@ extern char *rs6000_cpu_string;
 /* Width of wchar_t in bits.  */
 #define WCHAR_TYPE_SIZE 16
 
+/* A C expression for the size in bits of the type `short' on the
+   target machine.  If you don't define this, the default is half a
+   word.  (If this would be less than one storage unit, it is
+   rounded up to one unit.)  */
+#define SHORT_TYPE_SIZE 16
+
+/* A C expression for the size in bits of the type `int' on the
+   target machine.  If you don't define this, the default is one
+   word.  */
+#define INT_TYPE_SIZE (TARGET_64BIT ? 64 : 32)
+#define MAX_INT_TYPE_SIZE 64
+
+/* A C expression for the size in bits of the type `long' on the
+   target machine.  If you don't define this, the default is one
+   word.  */
+#define LONG_TYPE_SIZE (TARGET_64BIT ? 64 : 32)
+#define MAX_LONG_TYPE_SIZE 64
+
+/* A C expression for the size in bits of the type `long long' on the
+   target machine.  If you don't define this, the default is two
+   words.  */
+#define LONG_LONG_TYPE_SIZE 64
+
+/* A C expression for the size in bits of the type `char' on the
+   target machine.  If you don't define this, the default is one
+   quarter of a word.  (If this would be less than one storage unit,
+   it is rounded up to one unit.)  */
+#define CHAR_TYPE_SIZE BITS_PER_UNIT
+
+/* A C expression for the size in bits of the type `float' on the
+   target machine.  If you don't define this, the default is one
+   word.  */
+#define FLOAT_TYPE_SIZE 32
+
+/* A C expression for the size in bits of the type `double' on the
+   target machine.  If you don't define this, the default is two
+   words.  */
+#define DOUBLE_TYPE_SIZE 64
+
+/* A C expression for the size in bits of the type `long double' on
+   the target machine.  If you don't define this, the default is two
+   words.  */
+#define LONG_DOUBLE_TYPE_SIZE 64
+
 /* Width in bits of a pointer.
    See also the macro `Pmode' defined below.  */
-#define POINTER_SIZE 32
+#define POINTER_SIZE (TARGET_64BIT ? 64 : 32)
 
 /* Allocation boundary (in *bits*) for storing arguments in argument list.  */
-#define PARM_BOUNDARY 32
+#define PARM_BOUNDARY (TARGET_64BIT ? 64 : 32)
 
 /* Boundary (in *bits*) on which stack pointer should be aligned.  */
 #define STACK_BOUNDARY 64
@@ -305,7 +404,7 @@ extern char *rs6000_cpu_string;
 #define FUNCTION_BOUNDARY 32
 
 /* No data type wants to be aligned rounder than this.  */
-#define BIGGEST_ALIGNMENT 32
+#define BIGGEST_ALIGNMENT (TARGET_64BIT ? 64 : 32)
 
 /* Alignment of field after `int : 0' in a structure.  */
 #define EMPTY_FIELD_BOUNDARY 32
@@ -355,10 +454,16 @@ extern char *rs6000_cpu_string;
 
    On RS/6000, r1 is used for the stack and r2 is used as the TOC pointer.  
 
-   cr5 is not supposed to be used.  */
+   cr5 is not supposed to be used.
+
+   On System V implementations, r13 is fixed and not available for use.  */
+
+#ifndef FIXED_R13
+#define FIXED_R13 0
+#endif
 
 #define FIXED_REGISTERS  \
-  {0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
+  {0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, FIXED_R13, 0, 0, \
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
@@ -372,7 +477,7 @@ extern char *rs6000_cpu_string;
    Aside from that, you can include as many other registers as you like.  */
 
 #define CALL_USED_REGISTERS  \
-  {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, \
+  {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, FIXED_R13, 0, 0, \
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, \
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
@@ -434,7 +539,7 @@ extern char *rs6000_cpu_string;
 
 #define HARD_REGNO_NREGS(REGNO, MODE)   \
   (FP_REGNO_P (REGNO)			\
-   ? ((GET_MODE_SIZE (MODE) + 2 * UNITS_PER_WORD - 1) / (2 * UNITS_PER_WORD)) \
+   ? ((GET_MODE_SIZE (MODE) + UNITS_PER_FP_WORD - 1) / UNITS_PER_FP_WORD) \
    : ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD))
 
 /* Value is 1 if hard register REGNO can hold a value of machine-mode MODE.
@@ -447,7 +552,7 @@ extern char *rs6000_cpu_string;
   (FP_REGNO_P (REGNO) ?						\
    (GET_MODE_CLASS (MODE) == MODE_FLOAT				\
     || (GET_MODE_CLASS (MODE) == MODE_INT			\
-	&& GET_MODE_SIZE (MODE) == 2 * UNITS_PER_WORD))		\
+	&& GET_MODE_SIZE (MODE) == UNITS_PER_FP_WORD))		\
    : CR_REGNO_P (REGNO) ? GET_MODE_CLASS (MODE) == MODE_CC	\
    : ! INT_REGNO_P (REGNO) ? (GET_MODE_CLASS (MODE) == MODE_INT	\
 			      && GET_MODE_SIZE (MODE) <= UNITS_PER_WORD) \
@@ -479,9 +584,11 @@ extern char *rs6000_cpu_string;
    : (CLASS1) == FLOAT_REGS && (CLASS2) != FLOAT_REGS ? 10	\
    : (CLASS1) != FLOAT_REGS && (CLASS2) == FLOAT_REGS ? 10	\
    : (((CLASS1) == SPECIAL_REGS || (CLASS1) == MQ_REGS		\
-       || (CLASS1) == LINK_REGS || (CLASS1) == CTR_REGS)	\
+       || (CLASS1) == LINK_REGS || (CLASS1) == CTR_REGS		\
+       || (CLASS1) == LINK_OR_CTR_REGS)				\
       && ((CLASS2) == SPECIAL_REGS || (CLASS2) == MQ_REGS	\
-	  || (CLASS2) == LINK_REGS || (CLASS2) == CTR_REGS)) ? 10 \
+	  || (CLASS2) == LINK_REGS || (CLASS2) == CTR_REGS	\
+	  || (CLASS2) == LINK_OR_CTR_REGS)) ? 10 \
    : 2)
 
 /* A C expressions returning the cost of moving data of MODE from a register to
@@ -515,11 +622,17 @@ extern char *rs6000_cpu_string;
 
 /* Define this macro to change register usage conditional on target flags.
    Set MQ register fixed (already call_used) if not POWER architecture
-   (RIOS1, RIOS2, RSC, and PPC601) so that it will not be allocated.  */
+   (RIOS1, RIOS2, RSC, and PPC601) so that it will not be allocated.
+   Conditionally disable FPRs.  */
 
-#define CONDITIONAL_REGISTER_USAGE					\
-    if (!TARGET_POWER)							\
-	fixed_regs[64] = 1;
+#define CONDITIONAL_REGISTER_USAGE	\
+{					\
+  if (! TARGET_POWER)			\
+    fixed_regs[64] = 1;			\
+  if (TARGET_SOFT_FLOAT)		\
+    for (i = 32; i < 64; i++)		\
+      fixed_regs[i] = call_used_regs[i] = 1; \
+}
 
 /* Specify the registers used for certain standard purposes.
    The values of these macros are register numbers.  */
@@ -682,6 +795,7 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
 
 #define EXTRA_CONSTRAINT(OP, C)						\
   ((C) == 'Q' ? GET_CODE (OP) == MEM && GET_CODE (XEXP (OP, 0)) == REG	\
+   : (C) == 'R' ? LEGITIMATE_CONSTANT_POOL_ADDRESS_P (OP)		\
    : 0)
 
 /* Given an rtx X being reloaded into a reg required to be
@@ -717,7 +831,7 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
    except in the FP regs, where a single reg is enough for two words.  */
 #define CLASS_MAX_NREGS(CLASS, MODE)	\
  ((CLASS) == FLOAT_REGS			\
-  ? ((GET_MODE_SIZE (MODE) + 2 * UNITS_PER_WORD - 1) / (2 * UNITS_PER_WORD)) \
+  ? ((GET_MODE_SIZE (MODE) + UNITS_PER_FP_WORD - 1) / UNITS_PER_FP_WORD) \
   : ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD))
 
 /* If defined, gives a class of registers that cannot be used as the
@@ -749,7 +863,8 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
    except for dynamic allocations.  So we start after the fixed area and
    outgoing parameter area.  */
 
-#define STARTING_FRAME_OFFSET (current_function_outgoing_args_size + 24)
+#define STARTING_FRAME_OFFSET (current_function_outgoing_args_size \
+			       + (TARGET_64BIT ? 48 : 24))
 
 /* If we generate an insn to push BYTES bytes,
    this says how many the stack pointer really advances by.
@@ -759,12 +874,12 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
 /* Offset of first parameter from the argument pointer register value.
    On the RS/6000, we define the argument pointer to the start of the fixed
    area.  */
-#define FIRST_PARM_OFFSET(FNDECL) 24
+#define FIRST_PARM_OFFSET(FNDECL) (TARGET_64BIT ? 48 : 24)
 
 /* Define this if stack space is still allocated for a parameter passed
    in a register.  The value is the number of bytes allocated to this
    area.  */
-#define REG_PARM_STACK_SPACE(FNDECL)	32
+#define REG_PARM_STACK_SPACE(FNDECL)	(TARGET_64BIT ? 64 : 32)
 
 /* Define this if the above stack space is to be considered part of the
    space allocated by the caller.  */
@@ -773,7 +888,7 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
 /* This is the difference between the logical top of stack and the actual sp.
 
    For the RS/6000, sp points past the fixed area. */
-#define STACK_POINTER_OFFSET 24
+#define STACK_POINTER_OFFSET (TARGET_64BIT ? 48 : 24)
 
 /* Define this if the maximum size of all the outgoing args is to be
    accumulated and pushed during the prologue.  The amount can be
@@ -782,11 +897,12 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
 
 /* Value is the number of bytes of arguments automatically
    popped when returning from a subroutine call.
+   FUNDECL is the declaration node of the function (as a tree),
    FUNTYPE is the data type of the function (as a tree),
    or for a library call it is an identifier node for the subroutine name.
    SIZE is the number of bytes of arguments passed on the stack.  */
 
-#define RETURN_POPS_ARGS(FUNTYPE,SIZE) 0
+#define RETURN_POPS_ARGS(FUNDECL,FUNTYPE,SIZE) 0
 
 /* Define how to find the value returned by a function.
    VALTYPE is the data type of the value (as a tree).
@@ -794,17 +910,17 @@ enum reg_class { NO_REGS, BASE_REGS, GENERAL_REGS, FLOAT_REGS,
    otherwise, FUNC is 0.
 
    On RS/6000 an integer value is in r3 and a floating-point value is in 
-   fp1.  */
+   fp1, unless -msoft-float.  */
 
 #define FUNCTION_VALUE(VALTYPE, FUNC)	\
   gen_rtx (REG, TYPE_MODE (VALTYPE),	\
-	   TREE_CODE (VALTYPE) == REAL_TYPE ? 33 : 3)
+	   TREE_CODE (VALTYPE) == REAL_TYPE && TARGET_HARD_FLOAT ? 33 : 3)
 
 /* Define how to find the value returned by a library function
    assuming the value has mode MODE.  */
 
 #define LIBCALL_VALUE(MODE)		\
-  gen_rtx (REG, MODE, GET_MODE_CLASS (MODE) == MODE_FLOAT ? 33 : 3)
+  gen_rtx (REG, MODE, GET_MODE_CLASS (MODE) == MODE_FLOAT && TARGET_HARD_FLOAT ? 33 : 3)
 
 /* The definition of this macro implies that there are cases where
    a scalar value cannot be returned in registers.
@@ -887,7 +1003,7 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
 
 /* Non-zero if we can use a floating-point register to pass this arg.  */
 #define USE_FP_FOR_ARG_P(CUM,MODE,TYPE)	\
-  (GET_MODE_CLASS (MODE) == MODE_FLOAT && (CUM).fregno < 46)
+  (GET_MODE_CLASS (MODE) == MODE_FLOAT && (CUM).fregno < 46 && TARGET_HARD_FLOAT)
 
 /* Determine where to put an argument to a function.
    Value is zero to push the argument on the stack,
@@ -1047,6 +1163,27 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
 		  force_reg (SImode, (CXT)));			\
 }
 
+/* Definitions for __builtin_return_address and __builtin_frame_address.
+   __builtin_return_address (0) should give link register (65), enable
+   this. */
+/* This should be uncommented, so that the link register is used, but
+   currently this would result in unmatched insns and spilling fixed
+   registers so we'll leave it for another day.  When these problems are
+   taken care of one additional fetch will be necessary in RETURN_ADDR_RTX.
+   (mrs) */
+/* #define RETURN_ADDR_IN_PREVIOUS_FRAME */
+/* The current return address is in link register (65).  The return address
+   of anything farther back is accessed normally at an offset of 8 from the
+   frame pointer.  */
+#define RETURN_ADDR_RTX(count, frame)			\
+  ((count == -1)					\
+   ? gen_rtx (REG, Pmode, 65)				\
+   : copy_to_reg (gen_rtx (MEM, Pmode,			\
+			   memory_address (Pmode, 	\
+					   plus_constant (copy_to_reg (gen_rtx (MEM, Pmode, \
+										memory_address (Pmode, frame))), \
+							  8)))))
+
 /* Definitions for register eliminations.
 
    We have two registers that can be eliminated on the RS/6000.  First, the
@@ -1259,6 +1396,7 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
   if (LEGITIMATE_OFFSET_ADDRESS_P (MODE, X))		\
     goto ADDR;						\
   if ((MODE) != DImode && (MODE) != TImode		\
+      && (TARGET_HARD_FLOAT || (MODE) != DFmode)	\
       && LEGITIMATE_INDEXED_ADDRESS_P (X))		\
     goto ADDR;						\
 }
@@ -1305,6 +1443,7 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
     }								\
   else if (GET_CODE (X) == PLUS && GET_CODE (XEXP (X, 0)) == REG \
 	   && GET_CODE (XEXP (X, 1)) != CONST_INT		\
+	   && (TARGET_HARD_FLOAT || (MODE) != DFmode)		\
 	   && (MODE) != DImode && (MODE) != TImode) 		\
     {								\
       (X) = gen_rtx (PLUS, SImode, XEXP (X, 0),			\
@@ -1363,7 +1502,8 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
 
 /* Max number of bytes we can move from memory to memory
    in one reasonably fast instruction.  */
-#define MOVE_MAX 16
+#define MOVE_MAX (TARGET_POWERPC64 ? 8 : 4)
+#define MAX_MOVE_MAX 8
 
 /* Nonzero if access to memory by bytes is no faster than for words.
    Also non-zero if doing byte operations (specifically shifts) in registers
@@ -1410,12 +1550,12 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
 /* Specify the machine mode that pointers have.
    After generation of rtl, the compiler makes no further distinction
    between pointers and any other objects of this machine mode.  */
-#define Pmode SImode
+#define Pmode (TARGET_64BIT ? DImode : SImode)
 
 /* Mode of a function address in a call instruction (for indexing purposes).
 
    Doesn't matter on RS/6000.  */
-#define FUNCTION_MODE SImode
+#define FUNCTION_MODE (TARGET_64BIT ? DImode : SImode)
 
 /* Define this if addresses of constant functions
    shouldn't be put through pseudo regs where they can be cse'd.
@@ -1440,7 +1580,7 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
    of a switch statement.  If the code is computed here,
    return it with a return statement.  Otherwise, break from the switch.
 
-   On the RS/6000, if it is legal in the insn, it is free.  So this
+   On the RS/6000, if it is valid in the insn, it is free.  So this
    always returns 0.  */
 
 #define CONST_COSTS(RTX,CODE,OUTER_CODE) \
@@ -1468,6 +1608,7 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
     case PROCESSOR_PPC601:				\
     case PROCESSOR_PPC603:				\
       return COSTS_N_INSNS (5);				\
+    case PROCESSOR_PPC403:				\
     case PROCESSOR_PPC604:				\
     case PROCESSOR_PPC620:				\
       return COSTS_N_INSNS (4);				\
@@ -1486,6 +1627,8 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
       return COSTS_N_INSNS (19);			\
     case PROCESSOR_RIOS2:				\
       return COSTS_N_INSNS (13);			\
+    case PROCESSOR_PPC403:				\
+      return COSTS_N_INSNS (33);			\
     case PROCESSOR_PPC601:				\
       return COSTS_N_INSNS (36);			\
     case PROCESSOR_PPC603:				\
@@ -1494,6 +1637,8 @@ struct rs6000_args {int words, fregno, nargs_prototype; };
     case PROCESSOR_PPC620:				\
       return COSTS_N_INSNS (20);			\
     }							\
+  case FFS:						\
+    return COSTS_N_INSNS (4);				\
   case MEM:						\
     /* MEM should be slightly more expensive than (plus (reg) (const)) */ \
     return 5;
@@ -1557,6 +1702,19 @@ extern int rs6000_trunc_used;
 
 #define RS6000_ITRUNC "itrunc"
 #define RS6000_UITRUNC "uitrunc"
+
+/* Prefix and suffix to use to saving floating point */
+#ifndef SAVE_FP_PREFIX
+#define	SAVE_FP_PREFIX "._savef"
+#define SAVE_FP_SUFFIX ""
+#endif
+
+/* Prefix and suffix to use to restoring floating point */
+#ifndef RESTORE_FP_PREFIX
+#define	RESTORE_FP_PREFIX "._restf"
+#define RESTORE_FP_SUFFIX ""
+#endif
+
 
 /* Control the assembler format that we output.  */
 
@@ -1804,7 +1962,8 @@ toc_section ()						\
 {							\
   if ((TREE_CODE (EXP) == STRING_CST			\
        && !flag_writable_strings)			\
-      || (TREE_READONLY (EXP) && ! TREE_THIS_VOLATILE (EXP) \
+      || (TREE_CODE_CLASS (TREE_CODE (EXP)) == 'd'	\
+	  && TREE_READONLY (EXP) && ! TREE_THIS_VOLATILE (EXP) \
 	  && DECL_INITIAL (EXP)				\
 	  && (DECL_INITIAL (EXP) == error_mark_node	\
 	      || TREE_CONSTANT (DECL_INITIAL (EXP)))	\
@@ -1827,17 +1986,32 @@ toc_section ()						\
 /* This outputs NAME to FILE up to the first null or '['.  */
 
 #define RS6000_OUTPUT_BASENAME(FILE, NAME)	\
-  if ((NAME)[0] == '*' || (NAME)[strlen (NAME) - 1] != ']') \
-    assemble_name (FILE, NAME);  		\
-  else						\
-    {						\
-      int _len = strlen (NAME);			\
-      char *_p = alloca (_len + 1);		\
+  {						\
+    char *_p;					\
 						\
-      strcpy (_p, NAME);			\
-      _p[_len - 4] = '\0';			\
-      assemble_name (FILE, _p);			\
-    }
+    STRIP_NAME_ENCODING (_p, (NAME));		\
+    assemble_name ((FILE), _p);			\
+  }
+
+/* Remove any trailing [DS] or the like from the symbol name.  */
+
+#define STRIP_NAME_ENCODING(VAR,NAME)				\
+  do								\
+    {								\
+      if ((NAME)[0] == '*')					\
+	(VAR) = (NAME)+1;					\
+      else if ((NAME)[strlen (NAME) - 1] != ']')		\
+	(VAR) = (NAME);						\
+      else							\
+	{							\
+          int _len = strlen (NAME);				\
+	  (VAR) = alloca (_len + 1);				\
+								\
+          strcpy ((VAR), NAME);					\
+          (VAR)[_len - 4] = '\0';				\
+        }							\
+    }								\
+  while (0)
 
 /* Output something to declare an external symbol to the assembler.  Most
    assemblers don't need this.  
@@ -2155,7 +2329,7 @@ toc_section ()						\
 
 /* Define which CODE values are valid.  */
 
-#define PRINT_OPERAND_PUNCT_VALID_P(CODE)  ((CODE) == '.')
+#define PRINT_OPERAND_PUNCT_VALID_P(CODE)  ((CODE) == '.' || (CODE) == '*')
 
 /* Print a memory address as an operand to reference that memory location.  */
 
@@ -2175,6 +2349,8 @@ toc_section ()						\
   {"reg_or_cint_operand", {SUBREG, REG, CONST_INT}},		\
   {"easy_fp_constant", {CONST_DOUBLE}},				\
   {"reg_or_mem_operand", {SUBREG, MEM, REG}},			\
+  {"lwa_operand", {SUBREG, MEM, REG}},				\
+  {"low_32_bit_operand", {CONST_DOUBLE, CONST_INT}},		\
   {"fp_reg_or_mem_operand", {SUBREG, MEM, REG}},		\
   {"mem_or_easy_const_operand", {SUBREG, MEM, CONST_DOUBLE}},	\
   {"add_operand", {SUBREG, REG, CONST_INT}},			\
@@ -2186,10 +2362,69 @@ toc_section ()						\
   {"mask_operand", {CONST_INT}},				\
   {"call_operand", {SYMBOL_REF, REG}},				\
   {"current_file_function_operand", {SYMBOL_REF}},		\
-  {"input_operand", {SUBREG, MEM, REG, CONST_INT}},		\
+  {"input_operand", {SUBREG, MEM, REG, CONST_INT, SYMBOL_REF}},	\
   {"load_multiple_operation", {PARALLEL}},			\
   {"store_multiple_operation", {PARALLEL}},			\
   {"branch_comparison_operator", {EQ, NE, LE, LT, GE,		\
 				  GT, LEU, LTU, GEU, GTU}},	\
   {"scc_comparison_operator", {EQ, NE, LE, LT, GE,		\
 			       GT, LEU, LTU, GEU, GTU}},
+
+/* Declare functions in rs6000.c */
+extern void rs6000_override_options ();
+extern struct rtx_def *rs6000_immed_double_const ();
+extern int direct_return ();
+extern int any_operand ();
+extern int short_cint_operand ();
+extern int u_short_cint_operand ();
+extern int non_short_cint_operand ();
+extern int gpc_reg_operand ();
+extern int cc_reg_operand ();
+extern int reg_or_short_operand ();
+extern int reg_or_neg_short_operand ();
+extern int reg_or_u_short_operand ();
+extern int reg_or_cint_operand ();
+extern int easy_fp_constant ();
+extern int low_32_bit_operand ();
+extern int fp_reg_or_mem_operand ();
+extern int mem_or_easy_const_operand ();
+extern int add_operand ();
+extern int non_add_cint_operand ();
+extern int logical_operand ();
+extern int non_logical_operand ();
+extern int mask_constant ();
+extern int mask_operand ();
+extern int and_operand ();
+extern int non_and_cint_operand ();
+extern int reg_or_mem_operand ();
+extern int lwa_operand ();
+extern int call_operand ();
+extern int current_file_function_operand ();
+extern int input_operand ();
+extern int expand_block_move ();
+extern int load_multiple_operation ();
+extern int store_multiple_operation ();
+extern int branch_comparison_operator ();
+extern int scc_comparison_operator ();
+extern int includes_lshift_p ();
+extern int includes_rshift_p ();
+extern int registers_ok_for_quad_peep ();
+extern int addrs_ok_for_quad_peep ();
+extern enum reg_class secondary_reload_class ();
+extern int ccr_bit ();
+extern void print_operand ();
+extern void print_operand_address ();
+extern int first_reg_to_save ();
+extern int first_fp_reg_to_save ();
+extern int must_save_cr ();
+extern int rs6000_sa_size ();
+extern int rs6000_makes_calls ();
+extern int rs6000_pushes_stack ();
+extern void svr4_traceback ();
+extern void output_prolog ();
+extern void output_epilog ();
+extern void output_toc ();
+extern void output_ascii ();
+extern void rs6000_gen_section_name ();
+extern void output_function_profiler ();
+extern int rs6000_adjust_cost ();
