@@ -18,14 +18,15 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 /* Supposedly the same as vanilla sparc svr4, except for the stuff below: */
 #include "sparc/sysv4.h"
 
 #undef CPP_PREDEFINES
 #define CPP_PREDEFINES \
- "-Dsun -Dsparc -Dunix -D__svr4__ \
+ "-Dsun -Dsparc -Dunix -D__svr4__ -D__SVR4 \
   -Asystem(unix) -Asystem(svr4) -Acpu(sparc) -Amachine(sparc)\
   -D__GCC_NEW_VARARGS__"
 
@@ -98,24 +99,17 @@ do {									\
 /* We don't use the standard svr4 STARTFILE_SPEC because it's wrong for us.
    We don't use the standard LIB_SPEC only because we don't yet support c++ */
 
-/* If we cannot find the GNU *crt*.o files in the STANDARD_STARTFILE_PREFIX
-   directory, our fallback strategy must be to look for these files instead
-   in the Sun C 2.0 directory.  */
-
-#undef MD_STARTFILE_PREFIX
-#define MD_STARTFILE_PREFIX "/opt/SUNWspro/SC2.0/"
-
 #undef STARTFILE_SPEC
 #define STARTFILE_SPEC "%{!shared: \
 			 %{!symbolic: \
 			  %{pg:crt1.o%s}%{!pg:%{p:mcrt1.o%s}%{!p:crt1.o%s}} \
-			  %{pg:gmon.o%s} \
-			  %{pg:crti.o%s}%{!pg:crti.o%s} \
-			  %{ansi:/usr/ccs/lib/values-Xc.o%s} \
-			  %{!ansi: \
-			   %{traditional:/usr/ccs/lib/values-Xt.o%s} \
-			   %{!traditional:/usr/ccs/lib/values-Xa.o%s}}}} \
-			  crtbegin.o%s"
+			  %{pg:gmon.o%s}}} \
+			%{pg:crti.o%s}%{!pg:crti.o%s} \
+			%{ansi:values-Xc.o%s} \
+			%{!ansi: \
+			 %{traditional:values-Xt.o%s} \
+			 %{!traditional:values-Xa.o%s}} \
+			crtbegin.o%s"
 
 /* ??? Note: in order for -compat-bsd to work fully,
    we must somehow arrange to fixincludes /usr/ucbinclude
@@ -123,10 +117,10 @@ do {									\
 
 #undef LIB_SPEC
 #define LIB_SPEC \
-  "%{compat-bsd:-lucb -lsocket -lnsl -lelf -laio} \
-   %{!shared:%{!symbolic:-lc}} \
-   crtend.o%s \
-   %{!shared:%{!symbolic:%{pg:crtn.o%s}%{!pg:crtn.o%s}}}"
+  "%{compat-bsd:-lucb -lsocket -lnsl -lelf -laio} %{!shared:%{!symbolic:-lc}}"
+
+#undef  ENDFILE_SPEC
+#define ENDFILE_SPEC "crtend.o%s %{pg:crtn.o%s}%{!pg:crtn.o%s}"
 
 /* This should be the same as in svr4.h, except with -R added.  */
 #undef LINK_SPEC
@@ -134,11 +128,11 @@ do {									\
   "%{h*} %{V} %{v:%{!V:-V}} \
    %{b} %{Wl,*:%*} \
    %{static:-dn -Bstatic} \
-   %{shared:-G -dy} \
-   %{symbolic:-Bsymbolic -G -dy} \
+   %{shared:-G -dy -z text %{!h*:%{o*:-h %*}}} \
+   %{symbolic:-Bsymbolic -G -dy -z text %{!h*:%{o*:-h %*}}} \
    %{G:-G} \
    %{YP,*} \
-   %{R*} \
+   %{R*} %{!static:%{!R*:%{L*:-R %*}}} \
    %{compat-bsd: \
      %{!YP,*:%{p:-Y P,/usr/ucblib:/usr/ccs/lib/libp:/usr/lib/libp:/usr/ccs/lib:/usr/lib} \
        %{!p:-Y P,/usr/ucblib:/usr/ccs/lib:/usr/lib}} \
