@@ -1,5 +1,5 @@
 /* Definitions of target machine for GNU compiler, for Sun SPARC.
-   Copyright (C) 1987, 1988, 1989, 1992, 1994 Free Software Foundation, Inc.
+   Copyright (C) 1987, 88, 89, 92, 94, 1995 Free Software Foundation, Inc.
    Contributed by Michael Tiemann (tiemann@cygnus.com).
    64 bit SPARC V9 support by Michael Tiemann, Jim Wilson, and Doug Evans,
    at Cygnus Support.
@@ -18,7 +18,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 /* Note that some other tm.h files include this one and then override
    many of the definitions that relate to assembler syntax.  */
@@ -60,16 +61,18 @@ extern enum arch_type sparc_arch_type;
    -Asystem(unix) -Asystem(bsd) -Acpu(sparc) -Amachine(sparc)"
 #endif
 
-#define LIB_SPEC "%{!p:%{!pg:-lc}}%{p:-lc_p}%{pg:-lc_p} %{g:-lg}"
+#define LIB_SPEC "%{!shared:%{!p:%{!pg:-lc}}%{p:-lc_p}%{pg:-lc_p} %{g:-lg}}"
 
 /* Provide required defaults for linker -e and -d switches.  */
 
 #define LINK_SPEC \
- "%{!nostdlib:%{!r*:%{!e*:-e start}}} -dc -dp %{static:-Bstatic} %{assert*}"
+ "%{!shared:%{!nostdlib:%{!r*:%{!e*:-e start}}} -dc -dp} %{static:-Bstatic} \
+  %{assert*} %{shared:-assert pure-text}"
 
 /* Special flags to the Sun-4 assembler when using pipe for input.  */
 
-#define ASM_SPEC " %| %{!pg:%{!p:%{fpic:-k} %{fPIC:-k}}}"
+#define ASM_SPEC \
+  " %| %{R} %{!pg:%{!p:%{fpic:-k} %{fPIC:-k}}} %{keep-local-as-symbols:-L}"
 
 /* Define macros to distinguish architectures.  */
 
@@ -129,11 +132,11 @@ void sparc_override_options ();
 	  }								\
 	flag_omit_frame_pointer = 0;					\
       }									\
-    SUBTARGET_OVERRIDE_OPTIONS						\
+    SUBTARGET_OVERRIDE_OPTIONS;						\
     sparc_override_options ();						\
   } while (0)
 
-/* This is meant to be redefined in the host dependent files */
+/* This is meant to be redefined in the host dependent files.  */
 #define SUBTARGET_OVERRIDE_OPTIONS
 
 /* These compiler options take an argument.  We ignore -target for now.  */
@@ -240,7 +243,7 @@ extern int target_flags;
    Hence, we assume the upper 32 bits of symbolic addresses are zero, and
    avoid generating %uhi and %ulo terms.
    Pointers are still 64 bits though!  This option is for v9 only.  */
-/* ??? This option is deprecated.  Try to use -mcode-model=medium-low.  */
+/* ??? This option is deprecated.  Try to use -mmedlow.  */
 #define MASK_ENV32 0x10000
 #define TARGET_ENV32 (target_flags & MASK_ENV32)
 
@@ -327,45 +330,14 @@ extern int target_flags;
 /*  {"ptr64", MASK_PTR64}, */		\
 /*  {"ptr32", -MASK_PTR64}, */		\
     {"stack-bias", MASK_STACK_BIAS},	\
-    {"no-stack-bias", -MASK_STACK_BIAS},
+    {"no-stack-bias", -MASK_STACK_BIAS}, \
+    {"medlow", -MASK_CODE_MODEL},	\
+    {"medlow", MASK_MEDLOW},		\
+    {"medany", -MASK_CODE_MODEL},	\
+    {"medany", MASK_MEDANY},
 #else
 #define V9_SWITCHES
 #endif
-
-/* This macro is similar to `TARGET_SWITCHES' but defines names of
-   command options that have values.  Its definition is an
-   initializer with a subgrouping for each command option.
-
-   Each subgrouping contains a string constant, that defines the
-   fixed part of the option name, and the address of a variable. 
-   The variable, type `char *', is set to the variable part of the
-   given option if the fixed part matches.  The actual option name
-   is made by prepending `-m' to the specified name.
-
-   Here is an example which defines `-mshort-data-NUMBER'.  If the
-   given option is `-mshort-data-512', the variable `m88k_short_data'
-   will be set to the string `"512"'.
-
-	extern char *m88k_short_data;
-	#define TARGET_OPTIONS { { "short-data-", &m88k_short_data } }  */
-
-/* For v9, two values of "code model" are currently supported.
-
-   medium-low
-	32 bit address space starting at 0
-
-   medium-anywhere
-	32 bit text segment starting at 0
-	32 bit data segment(s) starting anywhere (determined at link time)
-	MEDANY_BASE_REG points to the start
-*/
-
-extern char *sparc_code_model;
-
-#define TARGET_OPTIONS \
-{							\
-  { "code-model=",	&sparc_code_model	}	\
-}
 
 /* target machine storage layout */
 
@@ -399,7 +371,7 @@ extern char *sparc_code_model;
 
 /* Width of a word, in units (bytes).  */
 #define UNITS_PER_WORD		(TARGET_V9 ? 8 : 4)
-#define MAX_UNITS_PER_WORD	8
+#define MIN_UNITS_PER_WORD	4
 
 /* Now define the sizes of the C data types.  */
 
@@ -476,6 +448,15 @@ extern char *sparc_code_model;
    Put them in the data section.  This macro is only used in this file.  */
 #define MAX_TEXT_ALIGN 32
 
+/* This forces all variables and constants to the data section when PIC.
+   This is because the SunOS 4 shared library scheme thinks everything in
+   text is a function, and patches the address to point to a loader stub.  */
+/* This is defined to zero for every system which doesn't use the a.out object
+   file format.  */
+#ifndef SUNOS4_SHARED_LIBRARIES
+#define SUNOS4_SHARED_LIBRARIES 0
+#endif
+
 /* This is defined differently for v9 in a cover file.  */
 #define SELECT_SECTION(T,RELOC)						\
 {									\
@@ -486,20 +467,21 @@ extern char *sparc_code_model;
 	  && (DECL_INITIAL (T) == error_mark_node			\
 	      || TREE_CONSTANT (DECL_INITIAL (T)))			\
 	  && DECL_ALIGN (T) <= MAX_TEXT_ALIGN				\
-	  && ! (flag_pic && (RELOC)))					\
+	  && ! (flag_pic && ((RELOC) || SUNOS4_SHARED_LIBRARIES)))	\
 	text_section ();						\
       else								\
 	data_section ();						\
     }									\
   else if (TREE_CODE (T) == CONSTRUCTOR)				\
     {									\
-      if (flag_pic != 0 && (RELOC) != 0)				\
+      if (flag_pic && ((RELOC) || SUNOS4_SHARED_LIBRARIES))		\
 	data_section ();						\
     }									\
   else if (*tree_code_type[(int) TREE_CODE (T)] == 'c')			\
     {									\
       if ((TREE_CODE (T) == STRING_CST && flag_writable_strings)	\
-	  || TYPE_ALIGN (TREE_TYPE (T)) > MAX_TEXT_ALIGN)		\
+	  || TYPE_ALIGN (TREE_TYPE (T)) > MAX_TEXT_ALIGN		\
+	  || (flag_pic && ((RELOC) || SUNOS4_SHARED_LIBRARIES)))	\
 	data_section ();						\
       else								\
 	text_section ();						\
@@ -512,7 +494,7 @@ extern char *sparc_code_model;
 #define SELECT_RTX_SECTION(MODE, X)		\
 {						\
   if (GET_MODE_BITSIZE (MODE) <= MAX_TEXT_ALIGN \
-      && ! (flag_pic && symbolic_operand (X)))  \
+      && ! (flag_pic && (symbolic_operand (X) || SUNOS4_SHARED_LIBRARIES)))  \
     text_section ();				\
   else						\
     data_section ();				\
@@ -649,7 +631,7 @@ do								\
       {								\
 	fixed_regs[2] = 0;					\
 	fixed_regs[3] = 0;					\
-	fixed_regs[4] = 0;					\
+	fixed_regs[4] = TARGET_MEDANY != 0;			\
       }								\
     if (TARGET_FLAT)						\
       {								\
@@ -741,12 +723,14 @@ extern int sparc_mode_class[];
    Zero means the frame pointer need not be set up (and parms
    may be accessed via the stack pointer) in functions that seem suitable.
    This is computed in `reload', in reload1.c.
+   Used in flow.c, global.c, and reload1.c.
 
-   Used in flow.c, global.c, and reload1.c.  */
-extern int leaf_function;
-
+   Being a non-leaf function does not mean a frame pointer is needed in the
+   flat window model.  However, the debugger won't be able to backtrace through
+   us with out it.  */
 #define FRAME_POINTER_REQUIRED \
-  (TARGET_FRW ? (current_function_calls_alloca || current_function_varargs) \
+  (TARGET_FRW ? (current_function_calls_alloca || current_function_varargs \
+		 || !leaf_function_p ()) \
    : ! (leaf_function_p () && only_leaf_regs_used ()))
 
 /* C statement to store the difference between the frame pointer
@@ -778,8 +762,10 @@ extern int leaf_function;
 /* Sparc ABI says that quad-precision floats and all structures are returned
    in memory.
    For v9, all aggregates are returned in memory.  */
-#define RETURN_IN_MEMORY(TYPE) \
-  (TYPE_MODE (TYPE) == BLKmode || (! TARGET_V9 && TYPE_MODE (TYPE) == TFmode))
+#define RETURN_IN_MEMORY(TYPE)				\
+  (TYPE_MODE (TYPE) == BLKmode				\
+   || (! TARGET_V9 && (TYPE_MODE (TYPE) == TFmode	\
+		       || TYPE_MODE (TYPE) == TCmode)))
 
 /* Functions which return large structures get the address
    to place the wanted value at offset 64 from the frame.
@@ -823,15 +809,35 @@ extern int leaf_function;
    class that represents their union.  */
 
 /* The SPARC has two kinds of registers, general and floating point.
+
+   For v9 we must distinguish between the upper and lower floating point
+   registers because the upper ones can't hold SFmode values.
+   HARD_REGNO_MODE_OK won't help here because reload assumes that register(s)
+   satisfying a group need for a class will also satisfy a single need for
+   that class.  EXTRA_FP_REGS is a bit of a misnomer as it covers all 64 fp
+   regs.
+
+   It is important that one class contains all the general and all the standard
+   fp regs.  Otherwise find_reg() won't properly allocate int regs for moves,
+   because reg_class_record() will bias the selection in favor of fp regs,
+   because reg_class_subunion[GENERAL_REGS][FP_REGS] will yield FP_REGS,
+   because FP_REGS > GENERAL_REGS.
+
+   It is also important that one class contain all the general and all the
+   fp regs.  Otherwise when spilling a DFmode reg, it may be from EXTRA_FP_REGS
+   but find_reloads() may use class GENERAL_OR_FP_REGS. This will cause
+   allocate_reload_reg() to bypass it causing an abort because the compiler
+   thinks it doesn't have a spill reg when in fact it does.
+
    v9 also has 4 floating point condition code registers.  Since we don't
    have a class that is the union of FPCC_REGS with either of the others,
    it is important that it appear first.  Otherwise the compiler will die
    trying to compile _fixunsdfsi because fix_truncdfsi2 won't match its
    constraints.  */
-   /* ??? As an experiment for v9, we treat all fp regs similarily here.  */
 
 #ifdef SPARCV9
-enum reg_class { NO_REGS, FPCC_REGS, GENERAL_REGS, FP_REGS, GENERAL_OR_FP_REGS,
+enum reg_class { NO_REGS, FPCC_REGS, GENERAL_REGS, FP_REGS, EXTRA_FP_REGS,
+		 GENERAL_OR_FP_REGS, GENERAL_OR_EXTRA_FP_REGS,
 		 ALL_REGS, LIM_REG_CLASSES };
 #else
 enum reg_class { NO_REGS, GENERAL_REGS, FP_REGS, ALL_REGS, LIM_REG_CLASSES };
@@ -843,11 +849,11 @@ enum reg_class { NO_REGS, GENERAL_REGS, FP_REGS, ALL_REGS, LIM_REG_CLASSES };
 
 #ifdef SPARCV9
 #define REG_CLASS_NAMES \
-  {"NO_REGS", "FPCC_REGS", "GENERAL_REGS", "FP_REGS", "GENERAL_OR_FP_REGS", \
-   "ALL_REGS" }
+  { "NO_REGS", "FPCC_REGS", "GENERAL_REGS", "FP_REGS", "EXTRA_FP_REGS", \
+    "GENERAL_OR_FP_REGS", "GENERAL_OR_EXTRA_FP_REGS", "ALL_REGS" }
 #else
 #define REG_CLASS_NAMES \
-  {"NO_REGS", "GENERAL_REGS", "FP_REGS", "ALL_REGS" }
+  { "NO_REGS", "GENERAL_REGS", "FP_REGS", "ALL_REGS" }
 #endif
 
 /* Define which registers fit in which classes.
@@ -856,8 +862,9 @@ enum reg_class { NO_REGS, GENERAL_REGS, FP_REGS, ALL_REGS, LIM_REG_CLASSES };
 
 #ifdef SPARCV9
 #define REG_CLASS_CONTENTS \
-  {{0, 0, 0, 0}, {0, 0, 0, 0xf}, {-2, 0, 0, 0}, {0, -1, -1, 0}, \
-   {-2, -1, -1, 0}, {-2, -1, -1, 0xf}}
+  {{0, 0, 0, 0}, {0, 0, 0, 0xf}, {-2, 0, 0, 0}, \
+   {0, -1, 0, 0}, {0, -1, -1, 0}, {-2, -1, 0, 0}, {-2, -1, -1, 0}, \
+   {-2, -1, -1, 0xf}}
 #else
 #if 0 && defined (__GNUC__)
 #define REG_CLASS_CONTENTS {0LL, 0xfffffffeLL, 0xffffffff00000000LL, 0xfffffffffffffffeLL}
@@ -872,11 +879,12 @@ enum reg_class { NO_REGS, GENERAL_REGS, FP_REGS, ALL_REGS, LIM_REG_CLASSES };
    or could index an array.  */
 
 #ifdef SPARCV9
-#define REGNO_REG_CLASS(REGNO)				\
-  ((REGNO) == 0 ? NO_REGS				\
-   : ((REGNO) < 32 ? GENERAL_REGS			\
-      : ((REGNO) < 96 ? FP_REGS				\
-	 : FPCC_REGS)))
+#define REGNO_REG_CLASS(REGNO) \
+  ((REGNO) == 0 ? NO_REGS		\
+   : (REGNO) < 32 ? GENERAL_REGS	\
+   : (REGNO) < 64 ? FP_REGS		\
+   : (REGNO) < 96 ? EXTRA_FP_REGS	\
+   : FPCC_REGS)
 #else
 #define REGNO_REG_CLASS(REGNO) \
   ((REGNO) >= 32 ? FP_REGS : (REGNO) == 0 ? NO_REGS : GENERAL_REGS)
@@ -995,14 +1003,26 @@ extern char leaf_reg_remap[];
 #define INDEX_REG_CLASS GENERAL_REGS
 #define BASE_REG_CLASS GENERAL_REGS
 
+/* Local macro to handle the two v9 classes of FP regs.  */
+#ifdef SPARCV9
+#define FP_REG_CLASS_P(CLASS) ((CLASS) == FP_REGS || (CLASS) == EXTRA_FP_REGS)
+#else
+#define FP_REG_CLASS_P(CLASS) ((CLASS) == FP_REGS)
+#endif
+
 /* Get reg_class from a letter such as appears in the machine description.  */
 
 #ifdef SPARCV9
 #define REG_CLASS_FROM_LETTER(C) \
-  ((C) == 'f' ? FP_REGS : (C) == 'c' ? FPCC_REGS : NO_REGS)
+  ((C) == 'f' ? FP_REGS		\
+   : (C) == 'e' ? EXTRA_FP_REGS	\
+   : (C) == 'c' ? FPCC_REGS	\
+   : NO_REGS)
 #else
+/* Coerce v9's 'e' class to 'f', so we can use 'e' in the .md file for
+   v8 and v9.  */
 #define REG_CLASS_FROM_LETTER(C) \
-  ((C) == 'f' ? FP_REGS : NO_REGS)
+  ((C) == 'f' ? FP_REGS : (C) == 'e' ? FP_REGS : NO_REGS)
 #endif
 
 /* The letters I, J, K, L and M in a register constraint string
@@ -1040,7 +1060,7 @@ extern char leaf_reg_remap[];
    if an 'E' constraint fails to match it.  */
 #define PREFERRED_RELOAD_CLASS(X,CLASS)			\
   (CONSTANT_P (X)					\
-   && ((CLASS) == FP_REGS				\
+   && (FP_REG_CLASS_P (CLASS)				\
        || (GET_MODE_CLASS (GET_MODE (X)) == MODE_FLOAT	\
 	   && (HOST_FLOAT_FORMAT != IEEE_FLOAT_FORMAT	\
 	       || HOST_BITS_PER_INT != BITS_PER_WORD)))	\
@@ -1057,20 +1077,21 @@ extern char leaf_reg_remap[];
    a paradoxical subreg in a float/fix conversion insn.  */
 
 #define SECONDARY_INPUT_RELOAD_CLASS(CLASS, MODE, IN)		\
-  (((CLASS) == FP_REGS && ((MODE) == HImode || (MODE) == QImode)\
-      && (GET_CODE (IN) == MEM					\
-	  || ((GET_CODE (IN) == REG || GET_CODE (IN) == SUBREG)	\
-	      && true_regnum (IN) == -1))) ? GENERAL_REGS : NO_REGS)
+  ((FP_REG_CLASS_P (CLASS) && ((MODE) == HImode || (MODE) == QImode) \
+    && (GET_CODE (IN) == MEM					\
+	|| ((GET_CODE (IN) == REG || GET_CODE (IN) == SUBREG)	\
+	    && true_regnum (IN) == -1))) ? GENERAL_REGS : NO_REGS)
 
 #define SECONDARY_OUTPUT_RELOAD_CLASS(CLASS, MODE, IN)		\
-  ((CLASS) == FP_REGS && ((MODE) == HImode || (MODE) == QImode)	\
-   && (GET_CODE (IN) == MEM					\
-       || ((GET_CODE (IN) == REG || GET_CODE (IN) == SUBREG)	\
-	   && true_regnum (IN) == -1)) ? GENERAL_REGS : NO_REGS)
+  ((FP_REG_CLASS_P (CLASS) && ((MODE) == HImode || (MODE) == QImode) \
+    && (GET_CODE (IN) == MEM					\
+	|| ((GET_CODE (IN) == REG || GET_CODE (IN) == SUBREG)	\
+	    && true_regnum (IN) == -1))) ? GENERAL_REGS : NO_REGS)
 
 /* On SPARC it is not possible to directly move data between 
    GENERAL_REGS and FP_REGS.  */
-#define SECONDARY_MEMORY_NEEDED(CLASS1, CLASS2, MODE) ((CLASS1) != (CLASS2))
+#define SECONDARY_MEMORY_NEEDED(CLASS1, CLASS2, MODE) \
+  (FP_REG_CLASS_P (CLASS1) != FP_REG_CLASS_P (CLASS2))
 
 /* Return the stack location to use for secondary memory needed reloads.
    We want to use the reserved location just below the frame pointer.
@@ -1078,7 +1099,7 @@ extern char leaf_reg_remap[];
    if the frame size is zero.  */
 #define SECONDARY_MEMORY_NEEDED_RTX(MODE) \
   (get_frame_size () == 0						\
-   ? assign_stack_local (mode, GET_MODE_SIZE (mode), 0)			\
+   ? assign_stack_local (MODE, GET_MODE_SIZE (MODE), 0)			\
    : gen_rtx (MEM, MODE, gen_rtx (PLUS, Pmode, frame_pointer_rtx,	\
 				  GEN_INT (STARTING_FRAME_OFFSET))))
 
@@ -1087,18 +1108,18 @@ extern char leaf_reg_remap[];
    For v8 we copy the default definition.  */
 #define SECONDARY_MEMORY_NEEDED_MODE(MODE) \
   (TARGET_V9							\
-   ? (GET_MODE_BITSIZE (mode) < 32				\
-      ? mode_for_size (32, GET_MODE_CLASS (mode), 0)		\
+   ? (GET_MODE_BITSIZE (MODE) < 32				\
+      ? mode_for_size (32, GET_MODE_CLASS (MODE), 0)		\
       : MODE)							\
-   : (GET_MODE_BITSIZE (mode) < BITS_PER_WORD			\
-      ? mode_for_size (BITS_PER_WORD, GET_MODE_CLASS (mode), 0)	\
+   : (GET_MODE_BITSIZE (MODE) < BITS_PER_WORD			\
+      ? mode_for_size (BITS_PER_WORD, GET_MODE_CLASS (MODE), 0)	\
       : MODE))
 
 /* Return the maximum number of consecutive registers
    needed to represent mode MODE in a register of class CLASS.  */
 /* On SPARC, this is the size of MODE in words.  */
 #define CLASS_MAX_NREGS(CLASS, MODE)	\
-  ((CLASS) == FP_REGS ? (GET_MODE_SIZE (MODE) + 3) / 4 \
+  (FP_REG_CLASS_P (CLASS) ? (GET_MODE_SIZE (MODE) + 3) / 4 \
    : (GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD)
 
 /* Stack layout; function entry, exit and calling.  */
@@ -1158,11 +1179,12 @@ extern char leaf_reg_remap[];
 
 /* Value is the number of bytes of arguments automatically
    popped when returning from a subroutine call.
+   FUNDECL is the declaration node of the function (as a tree),
    FUNTYPE is the data type of the function (as a tree),
    or for a library call it is an identifier node for the subroutine name.
    SIZE is the number of bytes of arguments passed on the stack.  */
 
-#define RETURN_POPS_ARGS(FUNTYPE,SIZE) 0
+#define RETURN_POPS_ARGS(FUNDECL,FUNTYPE,SIZE) 0
 
 /* Some subroutine macros specific to this machine.
    When !TARGET_FPU, put float return values in the general registers,
@@ -1471,9 +1493,6 @@ extern struct rtx_def *gen_compare_reg ();
 /* This function handles all v9 scc insns */
 
 extern int gen_v9_scc ();
-
-/* ??? This is a hack until conditional move support is complete.  */
-#define HAVE_conditional_move (TARGET_V9)
 
 /* Generate the special assembly code needed to tell the assembler whatever
    it might need to know about the return value of a function.
@@ -1509,6 +1528,7 @@ do {									\
    to do a "save" insn.  The decision about whether or not
    to do this is made in regclass.c.  */
 
+extern int leaf_function;
 #define FUNCTION_PROLOGUE(FILE, SIZE)				\
   (TARGET_FRW ? sparc_flat_output_function_prologue (FILE, SIZE) \
    : output_function_prologue (FILE, SIZE, leaf_function))
@@ -1859,7 +1879,11 @@ extern struct rtx_def *sparc_builtin_saveregs ();
       register rtx op0 = XEXP (X, 0);			\
       register rtx op1 = XEXP (X, 1);			\
       if (RTX_OK_FOR_BASE_P (op0)			\
-	  && CONSTANT_P (op1))				\
+	  && CONSTANT_P (op1)				\
+	  /* We can't allow TFmode, because an offset	\
+	     greater than or equal to the alignment (8)	\
+	     may cause the LO_SUM to overflow.  */	\
+	  && MODE != TFmode)				\
 	goto ADDR;					\
     }							\
   else if (GET_CODE (X) == CONST_INT && SMALL_INT (X))	\
@@ -2066,9 +2090,9 @@ extern struct rtx_def *legitimize_pic_address ();
    SPARC ABI.  */
 #define ADDTF3_LIBCALL "_Q_add"
 #define SUBTF3_LIBCALL "_Q_sub"
+#define NEGTF2_LIBCALL "_Q_neg"
 #define MULTF3_LIBCALL "_Q_mul"
 #define DIVTF3_LIBCALL "_Q_div"
-#define SQRTTF2_LIBCALL "_Q_sqrt"
 #define FLOATSITF2_LIBCALL "_Q_itoq"
 #define FIX_TRUNCTFSI2_LIBCALL "_Q_qtoi"
 #define FIXUNS_TRUNCTFSI2_LIBCALL "_Q_qtou"
@@ -2083,6 +2107,45 @@ extern struct rtx_def *legitimize_pic_address ();
 #define LTTF2_LIBCALL "_Q_flt"
 #define LETF2_LIBCALL "_Q_fle"
 
+/* We can define the TFmode sqrt optab only if TARGET_FPU.  This is because
+   with soft-float, the SFmode and DFmode sqrt instructions will be absent,
+   and the compiler will notice and try to use the TFmode sqrt instruction
+   for calls to the builtin function sqrt, but this fails.  */
+#define INIT_TARGET_OPTABS						\
+  do {									\
+    add_optab->handlers[(int) TFmode].libfunc				\
+      = gen_rtx (SYMBOL_REF, Pmode, ADDTF3_LIBCALL);			\
+    sub_optab->handlers[(int) TFmode].libfunc				\
+      = gen_rtx (SYMBOL_REF, Pmode, SUBTF3_LIBCALL);			\
+    neg_optab->handlers[(int) TFmode].libfunc				\
+      = gen_rtx (SYMBOL_REF, Pmode, NEGTF2_LIBCALL);			\
+    smul_optab->handlers[(int) TFmode].libfunc				\
+      = gen_rtx (SYMBOL_REF, Pmode, MULTF3_LIBCALL);			\
+    flodiv_optab->handlers[(int) TFmode].libfunc			\
+      = gen_rtx (SYMBOL_REF, Pmode, DIVTF3_LIBCALL);			\
+    eqtf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, EQTF2_LIBCALL);		\
+    netf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, NETF2_LIBCALL);		\
+    gttf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, GTTF2_LIBCALL);		\
+    getf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, GETF2_LIBCALL);		\
+    lttf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, LTTF2_LIBCALL);		\
+    letf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, LETF2_LIBCALL);		\
+    trunctfsf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, TRUNCTFSF2_LIBCALL);   \
+    trunctfdf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, TRUNCTFDF2_LIBCALL);   \
+    extendsftf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, EXTENDSFTF2_LIBCALL); \
+    extenddftf2_libfunc = gen_rtx (SYMBOL_REF, Pmode, EXTENDDFTF2_LIBCALL); \
+    floatsitf_libfunc = gen_rtx (SYMBOL_REF, Pmode, FLOATSITF2_LIBCALL);    \
+    fixtfsi_libfunc = gen_rtx (SYMBOL_REF, Pmode, FIX_TRUNCTFSI2_LIBCALL);  \
+    fixunstfsi_libfunc							\
+      = gen_rtx (SYMBOL_REF, Pmode, FIXUNS_TRUNCTFSI2_LIBCALL);		\
+    if (TARGET_FPU)							\
+      sqrt_optab->handlers[(int) TFmode].libfunc			\
+	= gen_rtx (SYMBOL_REF, Pmode, "_Q_sqrt");			\
+    INIT_SUBTARGET_OPTABS;						\
+  } while (0)
+
+/* This is meant to be redefined in the host dependent files */
+#define INIT_SUBTARGET_OPTABS
+
 /* Compute the cost of computing a constant rtl expression RTX
    whose rtx-code is CODE.  The body of this macro is a portion
    of a switch statement.  If the code is computed here,
@@ -2118,8 +2181,8 @@ extern struct rtx_def *legitimize_pic_address ();
    and another.
    ??? v9: We ignore FPCC_REGS on the assumption they'll never be seen.  */
 #define REGISTER_MOVE_COST(CLASS1, CLASS2) \
-  (((CLASS1 == FP_REGS && CLASS2 == GENERAL_REGS) \
-    || (CLASS1 == GENERAL_REGS && CLASS2 == FP_REGS)) ? 6 : 2)
+  (((FP_REG_CLASS_P (CLASS1) && (CLASS2) == GENERAL_REGS) \
+    || ((CLASS1) == GENERAL_REGS && FP_REG_CLASS_P (CLASS2))) ? 6 : 2)
 
 /* Provide the costs of a rtl expression.  This is in the body of a
    switch on CODE.  The purpose for the cost of MULT is to encourage
