@@ -1,5 +1,5 @@
 /* Definitions of target machine for GNU compiler.  Vax version.
-   Copyright (C) 1987, 1988, 1991, 1993 Free Software Foundation, Inc.
+   Copyright (C) 1987, 88, 91, 93, 94, 95 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -15,12 +15,13 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 
 /* Names to predefine in the preprocessor for this target machine.  */
 
-#define CPP_PREDEFINES "-Dvax -Dunix -Asystem(unix) -Asystem(bsd) -Acpu(vax) -Amachine(vax)"
+#define CPP_PREDEFINES "-Dvax -D__vax__ -Dunix -Asystem(unix) -Asystem(bsd) -Acpu(vax) -Amachine(vax)"
 
 /* If using g-format floating point, alter math.h.  */
 
@@ -352,13 +353,14 @@ gen_rtx (PLUS, Pmode, frame, gen_rtx (CONST_INT, VOIDmode, 12))
 
 /* Value is the number of bytes of arguments automatically
    popped when returning from a subroutine call.
+   FUNDECL is the declaration node of the function (as a tree),
    FUNTYPE is the data type of the function (as a tree),
    or for a library call it is an identifier node for the subroutine name.
    SIZE is the number of bytes of arguments passed on the stack.
 
    On the Vax, the RET insn always pops all the args for any function.  */
 
-#define RETURN_POPS_ARGS(FUNTYPE,SIZE) (SIZE)
+#define RETURN_POPS_ARGS(FUNDECL,FUNTYPE,SIZE) (SIZE)
 
 /* Define how to find the value returned by a function.
    VALTYPE is the data type of the value (as a tree).
@@ -884,7 +886,7 @@ gen_rtx (PLUS, Pmode, frame, gen_rtx (CONST_INT, VOIDmode, 12))
 
 #define RTX_COSTS(RTX,CODE,OUTER_CODE) case FIX: case FLOAT:	\
  case MULT: case DIV: case UDIV: case MOD: case UMOD:		\
- case LSHIFT: case ASHIFT: case LSHIFTRT: case ASHIFTRT:	\
+ case ASHIFT: case LSHIFTRT: case ASHIFTRT:			\
  case ROTATE: case ROTATERT: case PLUS: case MINUS: case IOR:	\
  case XOR: case AND: case NEG: case NOT: case ZERO_EXTRACT:	\
  case SIGN_EXTRACT: case MEM: return vax_rtx_cost(RTX)
@@ -912,7 +914,8 @@ gen_rtx (PLUS, Pmode, frame, gen_rtx (CONST_INT, VOIDmode, 12))
 /* note that it is very hard to accidentally create a number that fits in a
    double but not in a float, since their ranges are almost the same */
 
-#define CHECK_FLOAT_VALUE(mode, d) (check_float_value (mode, &d))
+#define CHECK_FLOAT_VALUE(MODE, D, OVERFLOW) \
+  ((OVERFLOW) = check_float_value (MODE, &D, OVERFLOW))
 
 /* For future reference:
    D Float: 9 bit, sign magnitude, excess 128 binary exponent
@@ -954,13 +957,15 @@ gen_rtx (PLUS, Pmode, frame, gen_rtx (CONST_INT, VOIDmode, 12))
 	   && GET_CODE (XVECEXP (EXP, 0, 0)) == SET)		\
     {								\
       if (GET_CODE (SET_SRC (XVECEXP (EXP, 0, 0))) == CALL)	\
-	CC_STATUS_INIT;						\
+	CC_STATUS_INIT;					        \
       else if (GET_CODE (SET_DEST (XVECEXP (EXP, 0, 0))) != PC) \
 	{ cc_status.flags = 0;					\
 	  cc_status.value1 = SET_DEST (XVECEXP (EXP, 0, 0));	\
-	  cc_status.value2 = SET_SRC (XVECEXP (EXP, 0, 0)); } } \
-  /* PARALLELs whose first element sets the PC are aob, sob insns.	\
-     They do change the cc's.  So drop through and forget the cc's.  */ \
+	  cc_status.value2 = SET_SRC (XVECEXP (EXP, 0, 0)); }   \
+      else							\
+	/* PARALLELs whose first element sets the PC are aob,   \
+	   sob insns.  They do change the cc's.  */		\
+	CC_STATUS_INIT; }					\
   else CC_STATUS_INIT;						\
   if (cc_status.value1 && GET_CODE (cc_status.value1) == REG	\
       && cc_status.value2					\
@@ -1172,6 +1177,17 @@ do { char dstr[30];							\
 ( (OUTPUT) = (char *) alloca (strlen ((NAME)) + 10),	\
   sprintf ((OUTPUT), "%s.%d", (NAME), (LABELNO)))
 
+/* When debugging, we want to output an extra dummy label so that gas
+   can distinguish between D_float and G_float prior to processing the
+   .stabs directive identifying type double.  */
+
+#define ASM_IDENTIFY_LANGUAGE(FILE)	\
+  do {								\
+    output_lang_identify (FILE);				\
+    if (write_symbols == DBX_DEBUG)				\
+      fprintf (FILE, "___vax_%c_doubles:\n", ASM_DOUBLE_CHAR);	\
+  } while (0)
+
 /* Define the parentheses used to group arithmetic operations
    in assembler code.  */
 
