@@ -1,5 +1,5 @@
 /* Configuration for GNU C-compiler for Vax.
-   Copyright (C) 1987 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1994, 1995 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -15,7 +15,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 /* #defines that need visibility everywhere.  */
 #define FALSE 0
@@ -48,8 +49,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #define HOST_FLOAT_FORMAT VAX_FLOAT_FORMAT
 
 #define SUCCESS_EXIT_CODE 1
-#define FATAL_EXIT_CODE (44 | 0x10000000)  /* Failure, and no DCL message.  */
-
+#define FATAL_EXIT_CODE (44 | 0x10000000)  /* Abort, and no DCL message.  */
 
 /* A couple of conditionals for execution machine are controlled here.  */
 #ifndef VMS
@@ -72,11 +72,11 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 /* Specify the list of include file directories.  */
 #define INCLUDE_DEFAULTS \
 {									\
-  { "GNU_GXX_INCLUDE:", 1},						\
-  { "GNU_CC_INCLUDE:", 0},	/* GNU includes */			\
-  { "SYS$SYSROOT:[SYSLIB.]", 0}, /* VAX-11 "C" includes */		\
-  { ".", 0},			/* Make normal VMS filespecs work.  */	\
-  { 0, 0}								\
+  { "GNU_GXX_INCLUDE:", 1, 1 },						\
+  { "GNU_CC_INCLUDE:", 0, 0 },	/* GNU includes */			\
+  { "SYS$SYSROOT:[SYSLIB.]", 0, 0 }, /* VAX-11 "C" includes */		\
+  { ".", 0, 1 },		/* Make normal VMS filespecs work.  */	\
+  { 0, 0, 0 }								\
 }
 
 /* Under VMS a directory specification can be enclosed either in square
@@ -155,6 +155,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #define current_function_uses_pic_offset_table	curfunc_uses_pic_offset_table
 #define expand_start_loop_continue_elsewhere	expnd_start_loop_cont_elsewhere
 #define flag_schedule_insns_after_reload	flag_sched_insns_after_reload
+#define lookup_name_current_level_global	lookup_name_current_level_gbl
 #define maybe_building_objc_message_expr	maybe_building_objc_msg_expr
 #define output_deferred_addressed_constants	output_deferred_addr_constants
 #define reg_overlap_mentioned_for_reload_p	reg_overlap_mtnd_for_reload_p
