@@ -17,7 +17,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 
 /* High-level class interface. */
@@ -27,6 +28,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include <stdio.h>
 #include "cp-tree.h"
 #include "class.h"
+#include "output.h"
 #include "flags.h"
 
 #include "obstack.h"
@@ -47,6 +49,7 @@ extern tree unary_complex_lvalue ();
 static struct harshness_code convert_harshness ();
 
 #define EVIL_RETURN(ARG)	((ARG).code = EVIL_CODE, (ARG))
+#define STD_RETURN(ARG)		((ARG).code = STD_CODE, (ARG))
 #define QUAL_RETURN(ARG)	((ARG).code = QUAL_CODE, (ARG))
 #define TRIVIAL_RETURN(ARG)	((ARG).code = TRIVIAL_CODE, (ARG))
 #define ZERO_RETURN(ARG)	((ARG).code = 0, (ARG))
@@ -117,6 +120,7 @@ convert_harshness (type, parmtype, parm)
   struct harshness_code h;
   register enum tree_code codel;
   register enum tree_code coder;
+  int lvalue;
 
   h.code = 0;
   h.distance = 0;
@@ -126,17 +130,22 @@ convert_harshness (type, parmtype, parm)
   n_convert_harshness++;
 #endif
 
-  if (TYPE_PTRMEMFUNC_P (type))
-    type = TYPE_PTRMEMFUNC_FN_TYPE (type);
-  if (TYPE_PTRMEMFUNC_P (parmtype))
-    parmtype = TYPE_PTRMEMFUNC_FN_TYPE (parmtype);
-
   if (TREE_CODE (parmtype) == REFERENCE_TYPE)
     {
       if (parm)
 	parm = convert_from_reference (parm);
       parmtype = TREE_TYPE (parmtype);
+      lvalue = 1;
     }
+  else if (parm)
+    lvalue = lvalue_p (parm);
+  else
+    lvalue = 0;
+
+  if (TYPE_PTRMEMFUNC_P (type))
+    type = TYPE_PTRMEMFUNC_FN_TYPE (type);
+  if (TYPE_PTRMEMFUNC_P (parmtype))
+    parmtype = TYPE_PTRMEMFUNC_FN_TYPE (parmtype);
 
   codel = TREE_CODE (type);
   coder = TREE_CODE (parmtype);
@@ -164,10 +173,36 @@ convert_harshness (type, parmtype, parm)
       if (coder != TREE_CODE (type))
 	return EVIL_RETURN (h);
 
+      if (type != parmtype && coder == METHOD_TYPE)
+	{
+	  tree ttl = TYPE_METHOD_BASETYPE (type);
+	  tree ttr = TYPE_METHOD_BASETYPE (parmtype);
+
+	  int b_or_d = get_base_distance (ttr, ttl, 0, 0);
+	  if (b_or_d < 0)
+	    {
+	      b_or_d = get_base_distance (ttl, ttr, 0, 0);
+	      if (b_or_d < 0)
+		return EVIL_RETURN (h);
+	      h.distance = -b_or_d;
+	    }
+	  else
+	    h.distance = b_or_d;
+	  h.code = STD_CODE;
+
+	  type = build_function_type
+	    (TREE_TYPE (type), TREE_CHAIN (TYPE_ARG_TYPES (type)));
+	  parmtype = build_function_type
+	    (TREE_TYPE (parmtype), TREE_CHAIN (TYPE_ARG_TYPES (parmtype)));
+	}
+
       /* We allow the default conversion between function type
 	 and pointer-to-function type for free.  */
       if (type == parmtype)
-	return ZERO_RETURN (h);
+	return h;
+
+      if (pedantic)
+	return EVIL_RETURN (h);
 
       /* Compare return types.  */
       p1 = TREE_TYPE (type);
@@ -200,14 +235,16 @@ convert_harshness (type, parmtype, parm)
 
 	  if (! BINFO_OFFSET_ZEROP (binfo))
 	    {
+#if 0
 	      static int explained = 0;
 	      if (h2.distance < 0)
-		message_2_types (sorry, "cannot cast `%d' to `%d' at function call site", p2, p1);
+		message_2_types (sorry, "cannot cast `%s' to `%s' at function call site", p2, p1);
 	      else
-		message_2_types (sorry, "cannot cast `%d' to `%d' at function call site", p1, p2);
+		message_2_types (sorry, "cannot cast `%s' to `%s' at function call site", p1, p2);
 
 	      if (! explained++)
 		sorry ("(because pointer values change during conversion)");
+#endif
 	      return EVIL_RETURN (h);
 	    }
 	}
@@ -259,28 +296,34 @@ convert_harshness (type, parmtype, parm)
     }
   else if (codel == POINTER_TYPE && coder == OFFSET_TYPE)
     {
+      tree ttl, ttr;
+
       /* Get to the OFFSET_TYPE that this might be.  */
       type = TREE_TYPE (type);
 
       if (coder != TREE_CODE (type))
 	return EVIL_RETURN (h);
 
-      if (TYPE_OFFSET_BASETYPE (type) == TYPE_OFFSET_BASETYPE (parmtype))
+      ttl = TYPE_OFFSET_BASETYPE (type);
+      ttr = TYPE_OFFSET_BASETYPE (parmtype);
+
+      if (ttl == ttr)
 	h.code = 0;
-      else if (UNIQUELY_DERIVED_FROM_P (TYPE_OFFSET_BASETYPE (type),
-			       TYPE_OFFSET_BASETYPE (parmtype)))
-	{
-	  h.code = STD_CODE;
-	  h.distance = 1;
-	}
-      else if (UNIQUELY_DERIVED_FROM_P (TYPE_OFFSET_BASETYPE (parmtype),
-			       TYPE_OFFSET_BASETYPE (type)))
-	{
-	  h.code = STD_CODE;
-	  h.distance = -1;
-	}
       else
-	return EVIL_RETURN (h);
+	{
+	  int b_or_d = get_base_distance (ttr, ttl, 0, 0);
+	  if (b_or_d < 0)
+	    {
+	      b_or_d = get_base_distance (ttl, ttr, 0, 0);
+	      if (b_or_d < 0)
+		return EVIL_RETURN (h);
+	      h.distance = -b_or_d;
+	    }
+	  else
+	    h.distance = b_or_d;
+	  h.code = STD_CODE;
+	}
+
       /* Now test the OFFSET_TYPE's target compatibility.  */
       type = TREE_TYPE (type);
       parmtype = TREE_TYPE (parmtype);
@@ -300,6 +343,21 @@ convert_harshness (type, parmtype, parm)
   if (coder == VOID_TYPE)
     return EVIL_RETURN (h);
 
+  if (codel == BOOLEAN_TYPE)
+    {
+      if (INTEGRAL_CODE_P (coder) || coder == REAL_TYPE)
+	return STD_RETURN (h);
+      else if (coder == POINTER_TYPE || coder == OFFSET_TYPE)
+	{
+	  /* Make this worse than any conversion to another pointer.
+	     FIXME this is how I think the language should work, but it may not
+	     end up being how the language is standardized (jason 1/30/95).  */
+	  h.distance = 32767;
+	  return STD_RETURN (h);
+	}
+      return EVIL_RETURN (h);
+    }
+
   if (INTEGRAL_CODE_P (codel))
     {
       /* Control equivalence of ints an enums.  */
@@ -375,11 +433,10 @@ convert_harshness (type, parmtype, parm)
       register tree ttr = TYPE_MAIN_VARIANT (TREE_TYPE (parmtype));
       int penalty = 4 * (ttl != ttr);
 
-      /* Anything converts to void *.  void * converts to anything.
-	 Since these may be `const void *' (etc.) use VOID_TYPE
-	 instead of void_type_node.  Otherwise, the targets must be the same,
-	 except that we do allow (at some cost) conversion between signed and
-	 unsigned pointer types.  */
+      /* Anything converts to void *.  Since this may be `const void *'
+	 (etc.) use VOID_TYPE instead of void_type_node.  Otherwise, the
+	 targets must be the same, except that we do allow (at some cost)
+	 conversion between signed and unsigned pointer types.  */
 
       if ((TREE_CODE (ttl) == METHOD_TYPE
 	   || TREE_CODE (ttl) == FUNCTION_TYPE)
@@ -396,7 +453,8 @@ convert_harshness (type, parmtype, parm)
 	}
 
 #if 1
-      if (TREE_CODE (ttl) != VOID_TYPE && TREE_CODE (ttr) != VOID_TYPE)
+      if (TREE_CODE (ttl) != VOID_TYPE
+	  && (TREE_CODE (ttr) != VOID_TYPE || !parm || !integer_zerop (parm)))
 	{
 	  if (TREE_UNSIGNED (ttl) != TREE_UNSIGNED (ttr))
 	    {
@@ -404,7 +462,7 @@ convert_harshness (type, parmtype, parm)
 	      ttr = unsigned_type (ttr);
 	      penalty = 10;
 	    }
-	  if (! comp_target_types (ttl, ttr, 0))
+	  if (comp_target_types (ttl, ttr, 0) <= 0)
 	    return EVIL_RETURN (h);
 	}
 #else
@@ -414,7 +472,7 @@ convert_harshness (type, parmtype, parm)
 		&& (ttl = unsigned_type (ttl),
 		    ttr = unsigned_type (ttr),
 		    penalty = 10, 0))
-	    || (comp_target_types (ttl, ttr, 0))))
+	    || (comp_target_types (ttl, ttr, 0) > 0)))
 	return EVIL_RETURN (h);
 #endif
 
@@ -503,14 +561,21 @@ convert_harshness (type, parmtype, parm)
 
       ttl = TREE_TYPE (type);
 
-      /* When passing a non-const argument into a const reference (or vice
-	 versa), dig it a little, so a non-const reference is preferred
-	 over this one. (mrs) */
-      if (TYPE_READONLY (ttl) != constp
-	  || TYPE_VOLATILE (ttl) != volatilep)
-	penalty = 2;
-      else
-	penalty = 0;
+      /* Only allow const reference binding if we were given a parm to deal
+         with, since it isn't really a conversion.  This is a hack to
+         prevent build_type_conversion from finding this conversion, but
+         still allow overloading to find it.  */
+      if (! lvalue && ! (parm && TYPE_READONLY (ttl)))
+	return EVIL_RETURN (h);
+
+      if (TYPE_READONLY (ttl) < constp
+	  || TYPE_VOLATILE (ttl) < volatilep)
+	return EVIL_RETURN (h);
+
+      /* When passing a non-const argument into a const reference, dig it a
+	 little, so a non-const reference is preferred over this one.  */
+      penalty = ((TYPE_READONLY (ttl) > constp)
+		 + (TYPE_VOLATILE (ttl) > volatilep));
 
       ttl = TYPE_MAIN_VARIANT (ttl);
 
@@ -520,98 +585,17 @@ convert_harshness (type, parmtype, parm)
 	  form = TREE_CODE (intype);
 	}
 
-      if (ttl == intype && penalty == 0)
-	return ZERO_RETURN (h);
-      else
-	penalty = 2;
-
-      if (TREE_UNSIGNED (ttl) ^ TREE_UNSIGNED (intype))
-	{
-	  ttl = unsigned_type (ttl);
-	  intype = unsigned_type (intype);
-	  penalty += 2;
-	}
-
       ttr = intype;
 
-      /* If the initializer is not an lvalue, then it does not
-	 matter if we make life easier for the programmer
-	 by creating a temporary variable with which to
-	 hold the result.  */
-      if (parm && (INTEGRAL_CODE_P (coder)
-		   || coder == REAL_TYPE)
-	  && ! lvalue_p (parm))
-	{
-	  h = convert_harshness (ttl, ttr, NULL_TREE);
-	  if (penalty > 2 || h.code != 0)
-	    h.code |= STD_CODE;
-	  else
-	    h.code |= TRIVIAL_CODE;
-	  h.distance = 0;
-	  return h;
-	}
+      /* Maybe handle conversion to base here?  */
 
-      if (ttl == ttr)
+      h = convert_harshness (ttl, ttr, NULL_TREE);
+      if (penalty && h.code == 0)
 	{
-	  if (penalty > 2)
-	    {
-	      h.code = STD_CODE;
-	      h.distance = 0;
-	    }
-	  else
-	    {
-	      h.code = TRIVIAL_CODE;
-	      /* We set this here so that build_overload_call_real will be
-		 able to see the penalty we found, rather than just looking
-		 at a TRIVIAL_CODE with no other information.  */
-	      h.int_penalty = penalty;
-	    }
-	  return h;
-	}
-
-      /* Pointers to voids always convert for pointers.  But
-	 make them less natural than more specific matches.  */
-      if (TREE_CODE (ttl) == POINTER_TYPE && TREE_CODE (ttr) == POINTER_TYPE)
-	{
-	  if (TREE_TYPE (ttl) == void_type_node
-	      || TREE_TYPE (ttr) == void_type_node)
-	    {
-	      h.code = STD_CODE;
-	      h.distance = 0;
-	      return h;
-	    }
-	}
-
-      /* Here it does matter.  If this conversion is from derived to base,
-	 allow it.  Otherwise, types must be compatible in the strong sense.  */
-      if (TREE_CODE (ttl) == RECORD_TYPE && TREE_CODE (ttr) == RECORD_TYPE)
-	{
-	  int b_or_d = get_base_distance (ttl, ttr, 0, 0);
-	  if (b_or_d < 0)
-	    {
-	      b_or_d = get_base_distance (ttr, ttl, 0, 0);
-	      if (b_or_d < 0)
-		return EVIL_RETURN (h);
-	      h.distance = -b_or_d;
-	    }
-	  /* Say that this conversion is relatively painless.
-	     If it turns out that there is a user-defined X(X&)
-	     constructor, then that will be invoked, but that's
-	     preferable to dealing with other user-defined conversions
-	     that may produce surprising results.  */
-	  else
-	    h.distance = b_or_d;
-	  h.code = STD_CODE;
-	  return h;
-	}
-
-      if (comp_target_types (ttl, intype, 1))
-	{
-	  if (penalty)
-	    h.code = STD_CODE;
-	  h.distance = 0;
-	  return h;
+	  h.code = QUAL_CODE;
+	  h.int_penalty = penalty;
 	}
+      return h;
     }
   if (codel == RECORD_TYPE && coder == RECORD_TYPE)
     {
@@ -631,6 +615,65 @@ convert_harshness (type, parmtype, parm)
   return EVIL_RETURN (h);
 }
 
+int
+user_harshness (type, parmtype, parm)
+     register tree type, parmtype;
+     tree parm;
+{
+  tree conv;
+  tree winner = NULL_TREE;
+  int code;
+
+  {
+    tree typename = build_typename_overload (type);
+    if (lookup_fnfields (TYPE_BINFO (parmtype), typename, 0))
+      return 0;
+  }
+			
+  for (conv = lookup_conversions (parmtype); conv; conv = TREE_CHAIN (conv))
+    {
+      struct harshness_code tmp;
+
+      if (winner && TREE_PURPOSE (winner) == TREE_PURPOSE (conv))
+	continue;
+
+      if (tmp = convert_harshness (type, TREE_VALUE (conv), NULL_TREE),
+	  tmp.code < USER_CODE)
+	{
+	  if (winner)
+	    return EVIL_CODE;
+	  else
+	    {
+	      winner = conv;
+	      code = tmp.code;
+	    }
+	}
+    }
+
+  if (winner)
+    return code;
+
+  return -1;
+}
+
+int
+can_convert (to, from)
+     tree to, from;
+{
+  struct harshness_code h;
+  h = convert_harshness (to, from, NULL_TREE);
+  return h.code < USER_CODE;
+}
+
+int
+can_convert_arg (to, from, arg)
+     tree to, from, arg;
+{
+  struct harshness_code h;
+  h = convert_harshness (to, from, arg);
+  return h.code < USER_CODE;
+}
+
 #ifdef DEBUG_MATCHING
 static char *
 print_harshness (h)
@@ -700,11 +743,17 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 
   int strike_index = 0, win;
   struct harshness_code lose;
+  extern int cp_silent;
 
 #ifdef GATHER_STATISTICS
   n_compute_conversion_costs++;
 #endif
 
+#ifndef DEBUG_MATCHING
+  /* We don't emit any warnings or errors while trying out each candidate.  */
+  cp_silent = 1;
+#endif
+
   cp->function = function;
   cp->arg = tta ? TREE_VALUE (tta) : NULL_TREE;
   cp->u.bad_arg = 0;		/* optimistic!  */
@@ -712,7 +761,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
   cp->h.code = 0;
   cp->h.distance = 0;
   cp->h.int_penalty = 0;
-  bzero (cp->harshness,
+  bzero ((char *) cp->harshness,
 	 (cp->h_len + 1) * sizeof (struct harshness_code));
 
   while (ttf && tta)
@@ -812,6 +861,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 	{
 	  cp->h.code = EVIL_CODE;
 	  cp->u.bad_arg = -1;
+	  cp_silent = 0;
 	  return;
 	}
       else
@@ -833,6 +883,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 	{
 	  cp->h.code = EVIL_CODE;
 	  cp->u.bad_arg = -2;
+	  cp_silent = 0;
 	  return;
 	}
       /* Store index of first default.  */
@@ -855,6 +906,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
       if (dont_convert_types)
 	{
 	  cp->h.code = EVIL_CODE;
+	  cp_silent = 0;
 	  return;
 	}
 
@@ -903,43 +955,14 @@ compute_conversion_costs (function, tta_in, cp, arglen)
 		  if (TYPE_LANG_SPECIFIC (actual_type)
 		      && TYPE_HAS_CONVERSION (actual_type))
 		    {
-		      tree conv;
-		      /* Don't issue warnings since we're only groping
-			 around for the right answer, we haven't yet
-			 committed to going with this solution.  */
-		      int old_inhibit_warnings = inhibit_warnings;
+		      int extra = user_harshness (formal_type, actual_type);
 
-		      inhibit_warnings = 1;
-		      conv = build_type_conversion
-			(CALL_EXPR, TREE_VALUE (ttf), TREE_VALUE (tta), 0);
-		      inhibit_warnings = old_inhibit_warnings;
-
-		      if (conv)
+		      if (extra == EVIL_CODE)
+			win += 2;
+		      else if (extra >= 0)
 			{
-			  if (conv == error_mark_node)
-			    win += 2;
-			  else
-			    {
-			      win++;
-			      if (TREE_CODE (conv) != CALL_EXPR)
-				extra_conversions = 1;
-			    }
-			}
-		      else if (TREE_CODE (TREE_VALUE (ttf)) == REFERENCE_TYPE)
-			{
-			  conv = build_type_conversion (CALL_EXPR, formal_type,
-							TREE_VALUE (tta), 0);
-			  if (conv)
-			    {
-			      if (conv == error_mark_node)
-				win += 2;
-			      else
-				{
-				  win++;
-				  if (TREE_CODE (conv) != CALL_EXPR)
-				    extra_conversions = 1;
-				}
-			    }
+			  win++;
+			  extra_conversions = extra;
 			}
 		    }
 		}
@@ -1002,6 +1025,7 @@ compute_conversion_costs (function, tta_in, cp, arglen)
     cp->h.code |= ELLIPSIS_CODE;
   if (user_strikes)
     cp->h.code |= USER_CODE;
+  cp_silent = 0;
 #ifdef DEBUG_MATCHING
   cp_error ("final eval %s", print_harshness (&cp->h));
 #endif
@@ -1357,6 +1381,7 @@ resolve_scope_to_name (outer_type, inner_stuff)
 
   if (outer_type == NULL_TREE)
     {
+      tree x;
       /* If we have something that's already a type by itself,
 	 use that.  */
       if (IDENTIFIER_HAS_TYPE_VALUE (inner_name))
@@ -1366,6 +1391,14 @@ resolve_scope_to_name (outer_type, inner_stuff)
 					  inner_type);
 	  return inner_name;
 	}
+      
+      x = lookup_name (inner_name, 0);
+
+      if (x && TREE_CODE (x) == NAMESPACE_DECL)
+	{
+	  x = lookup_namespace_name (x, inner_type);
+	  return x;
+	}
       return NULL_TREE;
     }
 
@@ -1428,11 +1461,11 @@ build_scoped_method_call (exp, scopes, name, parms)
       if (type != basetype)
 	cp_error ("type of `%E' does not match destructor type `%T' (type was `%T')",
 		  exp, basetype, type);
-      name = IDENTIFIER_TYPE_VALUE (TREE_OPERAND (name, 0));
-      if (basetype != name)
-	cp_error ("qualified type `%T' does not match destructor type `%T'",
+      name = TREE_OPERAND (name, 0);
+      if (basetype != get_type_value (name))
+	cp_error ("qualified type `%T' does not match destructor name `~%T'",
 		  basetype, name);
-      return void_zero_node;
+      return convert (void_type_node, exp);
     }
 
   if (! is_aggr_typedef (basename, 1))
@@ -1460,15 +1493,16 @@ build_scoped_method_call (exp, scopes, name, parms)
 	{
 	  /* Explicit call to destructor.  */
 	  name = TREE_OPERAND (name, 0);
-	  if (name != constructor_name (TREE_TYPE (decl)))
+	  if (! (name == constructor_name (TREE_TYPE (decl))
+		 || TREE_TYPE (decl) == get_type_value (name)))
 	    {
 	      cp_error
-		("qualified type `%T' does not match destructor type `%T'",
+		("qualified type `%T' does not match destructor name `~%T'",
 		 TREE_TYPE (decl), name);
 	      return error_mark_node;
 	    }
 	  if (! TYPE_HAS_DESTRUCTOR (TREE_TYPE (decl)))
-	    return void_zero_node;
+	    return convert (void_type_node, exp);
 	  
 	  return build_delete (TREE_TYPE (decl), decl, integer_two_node,
 			       LOOKUP_NORMAL|LOOKUP_NONVIRTUAL|LOOKUP_DESTRUCTOR,
@@ -1586,8 +1620,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
       result = build_method_call (instance, name, parms, basetype_path,
 				  (LOOKUP_SPECULATIVELY|flags)
 				  &~LOOKUP_COMPLAIN);
-      /* If it works, return it. */
-      if (result && result != error_mark_node)
+      /* If it finds a match, return it. */
+      if (result)
 	return build_method_call (instance, name, parms, basetype_path, flags);
       /* If it doesn't work, two argument delete must work */
       TREE_CHAIN (parms) = save_last;
@@ -1604,23 +1638,19 @@ build_method_call (instance, name, parms, basetype_path, flags)
       if (parms)
 	error ("destructors take no parameters");
       basetype = TREE_TYPE (instance);
-      if (IS_AGGR_TYPE (basetype))
+      if (TREE_CODE (basetype) == REFERENCE_TYPE)
+	basetype = TREE_TYPE (basetype);
+      if (! ((IS_AGGR_TYPE (basetype)
+	      && name == constructor_name (basetype))
+	     || basetype == get_type_value (name)))
 	{
-	  if (name == constructor_name (basetype))
-	    goto huzzah;
+	  cp_error ("destructor name `~%D' does not match type `%T' of expression",
+		    name, basetype);
+	  return convert (void_type_node, instance);
 	}
-      else
-	{
-	  if (basetype == get_type_value (name))
-	    goto huzzah;
-	}
-      cp_error ("destructor name `~%D' does not match type `%T' of expression",
-		name, basetype);
-      return void_zero_node;
 
-    huzzah:
       if (! TYPE_HAS_DESTRUCTOR (basetype))
-	return void_zero_node;
+	return convert (void_type_node, instance);
       instance = default_conversion (instance);
       instance_ptr = build_unary_op (ADDR_EXPR, instance, 0);
       return build_delete (build_pointer_type (basetype),
@@ -1718,10 +1748,24 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	need_vtbl = (dtor_label || ctor_label)
 	  ? unneeded : maybe_needed;
 
-      instance = C_C_D;
-      instance_ptr = current_class_decl;
-      result = build_field_call (TYPE_BINFO (current_class_type),
-				 instance_ptr, name, parms);
+      /* If `this' is a signature pointer and `name' is not a constructor,
+	 we are calling a signature member function.  In that case, set the
+	 `basetype' to the signature type and dereference the `optr' field.  */
+      if (IS_SIGNATURE_POINTER (basetype)
+	  && TYPE_IDENTIFIER (basetype) != name)
+	{
+	  basetype = SIGNATURE_TYPE (basetype);
+	  instance_ptr = build_optr_ref (instance);
+	  instance_ptr = convert (TYPE_POINTER_TO (basetype), instance_ptr);
+	  basetype_path = TYPE_BINFO (basetype);
+	}
+      else
+	{
+	  instance = C_C_D;
+	  instance_ptr = current_class_decl;
+	  basetype_path = TYPE_BINFO (current_class_type);
+	}
+      result = build_field_call (basetype_path, instance_ptr, name, parms);
 
       if (result)
 	return result;
@@ -1765,7 +1809,10 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	}
       else
 	{
-	  if (! IS_AGGR_TYPE (basetype))
+	  if (! IS_AGGR_TYPE (basetype)
+	      && ! (TYPE_LANG_SPECIFIC (basetype)
+		    && (IS_SIGNATURE_POINTER (basetype)
+			|| IS_SIGNATURE_REFERENCE (basetype))))
 	    goto non_aggr_error;
 
 	  /* If `instance' is a signature pointer/reference and `name' is
@@ -1795,7 +1842,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		 we can manage.  */
 	      tree temp = get_temp_name (TREE_TYPE (instance), 0);
 	      if (IS_AGGR_TYPE (TREE_TYPE (instance)))
-		expand_aggr_init (temp, instance, 0);
+		expand_aggr_init (temp, instance, 0, flags);
 	      else
 		{
 		  store_init_value (temp, instance);
@@ -1897,6 +1944,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	{
 	  TREE_VALUE (parm) = build_unary_op (ADDR_EXPR, TREE_VALUE (parm), 0);
 	}
+#if 0
+      /* This breaks reference-to-array parameters.  */
       if (TREE_CODE (t) == ARRAY_TYPE)
 	{
 	  /* Perform the conversion from ARRAY_TYPE to POINTER_TYPE in place.
@@ -1904,6 +1953,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  TREE_VALUE (parm) = default_conversion (TREE_VALUE (parm));
 	  t = TREE_TYPE (TREE_VALUE (parm));
 	}
+#endif
       if (t == error_mark_node)
 	return error_mark_node;
       last = build_tree_list (NULL_TREE, t);
@@ -1932,7 +1982,9 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	{
 	  constp = 0;
 	  volatilep = 0;
-	  parms = tree_cons (NULL_TREE, build1 (NOP_EXPR, TYPE_POINTER_TO (basetype), integer_zero_node), parms);
+	  instance_ptr = build_int_2 (0, 0);
+	  TREE_TYPE (instance_ptr) = TYPE_POINTER_TO (basetype);
+	  parms = tree_cons (NULL_TREE, instance_ptr, parms);
 	}
       else
 	{
@@ -1945,6 +1997,10 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  TREE_CALLS_NEW (instance_ptr) = 1;
 	  instance = build_indirect_ref (instance_ptr, NULL_PTR);
 
+#if 0
+	  /* This breaks initialization of a reference from a new
+             expression of a different type.  And it doesn't appear to
+             serve its original purpose any more, either.  jason 10/12/94 */
 	  /* If it's a default argument initialized from a ctor, what we get
 	     from instance_ptr will match the arglist for the FUNCTION_DECL
 	     of the constructor.  */
@@ -1953,6 +2009,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	      && TREE_CALLS_NEW (TREE_VALUE (TREE_OPERAND (TREE_VALUE (parms), 1))))
 	    parms = build_tree_list (NULL_TREE, instance_ptr);
 	  else
+#endif
 	    parms = tree_cons (NULL_TREE, instance_ptr, parms);
 	}
     }
@@ -2012,6 +2069,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
     return error_mark_node;
 
 
+#if 0
   /* Now, go look for this method name.  We do not find destructors here.
 
      Putting `void_list_node' on the end of the parmtypes
@@ -2021,6 +2079,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 				     1 + (name == constructor_name (save_basetype)
 					  || name == constructor_name_full (save_basetype)));
   TREE_CHAIN (last) = NULL_TREE;
+#endif
 
   for (pass = 0; pass < 2; pass++)
     {
@@ -2040,7 +2099,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  candidates
 	    = (struct candidate *) alloca ((ever_seen+1)
 					   * sizeof (struct candidate));
-	  bzero (candidates, (ever_seen + 1) * sizeof (struct candidate));
+	  bzero ((char *) candidates, (ever_seen + 1) * sizeof (struct candidate));
 	  cp = candidates;
 	  len = list_length (parms);
 	  ever_seen = 0;
@@ -2052,25 +2111,14 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	      tree parm = instance_ptr;
 
 	      if (TREE_CODE (TREE_TYPE (parm)) == REFERENCE_TYPE)
-		{
-		  /* TREE_VALUE (parms) may have been modified by now;
-                     restore it to its original value. */
-		  TREE_VALUE (parms) = parm;
-		  friend_parms = parms;
-		}
+		parm = convert_from_reference (parm);
 	      else if (TREE_CODE (TREE_TYPE (parm)) == POINTER_TYPE)
-		{
-		  tree new_type;
-		  parm = build_indirect_ref (parm, "friendifying parms (compiler error)");
-		  new_type = c_build_type_variant (TREE_TYPE (parm), constp,
-						   volatilep);
-		  new_type = build_reference_type (new_type);
-		  parm = convert (new_type, parm);
-		  friend_parms = tree_cons (NULL_TREE, parm, TREE_CHAIN (parms));
-		}
+		parm = build_indirect_ref (parm, "friendifying parms (compiler error)");
 	      else
 		my_friendly_abort (167);
 
+	      friend_parms = tree_cons (NULL_TREE, parm, TREE_CHAIN (parms));
+
 	      cp->h_len = len;
 	      cp->harshness = (struct harshness_code *)
 		alloca ((len + 1) * sizeof (struct harshness_code));
@@ -2111,6 +2159,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	    basetype_path = TREE_VALUE (basetype_path);
 	  basetype = BINFO_TYPE (basetype_path);
 
+#if 0
 	  /* Cast the instance variable if necessary.  */
 	  if (basetype != TYPE_MAIN_VARIANT
 	      (TREE_TYPE (TREE_TYPE (TREE_VALUE (parms)))))
@@ -2121,7 +2170,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		{
 		  tree type = build_pointer_type
 		    (build_type_variant (basetype, constp, volatilep));
-		  TREE_VALUE (parms) = convert_force (type, instance_ptr);
+		  TREE_VALUE (parms) = convert_force (type, instance_ptr, 0);
 		}
 	    }
 
@@ -2129,6 +2178,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	     the access-control rewrite will make this change more cleanly.  */
 	  if (TREE_VALUE (parms) == error_mark_node)
 	    return error_mark_node;
+#endif
 
 	  if (DESTRUCTOR_NAME_P (DECL_ASSEMBLER_NAME (function)))
 	    function = DECL_CHAIN (function);
@@ -2147,9 +2197,11 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		  && ! DECL_STATIC_FUNCTION_P (function))
 		continue;
 
+#if 0
 	      if (pass == 0
 		  && DECL_ASSEMBLER_NAME (function) == method_name)
 		goto found;
+#endif
 
 	      if (pass > 0)
 		{
@@ -2172,24 +2224,17 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		      cp->function = function;
 		      cp->basetypes = basetype_path;
 
+		      /* Don't allow non-converting constructors to convert. */
+		      if (flags & LOOKUP_ONLYCONVERTING
+			  && DECL_LANG_SPECIFIC (function)
+			  && DECL_NONCONVERTING_P (function))
+			continue;
+
 		      /* No "two-level" conversions.  */
 		      if (flags & LOOKUP_NO_CONVERSION
 			  && (cp->h.code & USER_CODE))
 			continue;
 
-		      /* If we used default parameters, we must
-			 check to see whether anyone else might
-			 use them also, and report a possible
-			 ambiguity.  */
-		      if (! TYPE_USES_MULTIPLE_INHERITANCE (save_basetype)
-			  && cp->harshness[len].distance == 0
-			  && cp->h.code < best)
-			{
-			  if (! DECL_STATIC_FUNCTION_P (function))
-			    TREE_VALUE (parms) = cp->arg;
-			  if (best == 1)
-			    goto found_and_maybe_warn;
-			}
 		      cp++;
 		    }
 		}
@@ -2244,6 +2289,7 @@ build_method_call (instance, name, parms, basetype_path, flags)
 	  if (cp - candidates > 1)
 	    {
 	      int n_candidates = cp - candidates;
+	      extern int warn_synth;
 	      TREE_VALUE (parms) = instance_ptr;
 	      cp = ideal_candidate (save_basetype, candidates,
 				    n_candidates, parms, len);
@@ -2251,14 +2297,25 @@ build_method_call (instance, name, parms, basetype_path, flags)
 		{
 		  if (flags & LOOKUP_COMPLAIN)
 		    {
-		      cp_error ("call of overloaded %s `%D' is ambiguous",
-				name_kind, name);
+		      TREE_CHAIN (last) = void_list_node;
+		      cp_error ("call of overloaded %s `%D(%A)' is ambiguous",
+				name_kind, name, TREE_CHAIN (parmtypes));
 		      print_n_candidates (candidates, n_candidates);
 		    }
 		  return error_mark_node;
 		}
 	      if (cp->h.code & EVIL_CODE)
 		return error_mark_node;
+	      if (warn_synth
+		  && DECL_NAME (cp->function) == ansi_opname[MODIFY_EXPR]
+		  && DECL_ARTIFICIAL (cp->function)
+		  && n_candidates == 2)
+		{
+		  cp_warning ("using synthesized `%#D' for copy assignment",
+			      cp->function);
+		  cp_warning_at ("  where cfront would use `%#D'",
+				 candidates->function);
+		}
 	    }
 	  else if (cp[-1].h.code & EVIL_CODE)
 	    {
@@ -2397,6 +2454,19 @@ build_method_call (instance, name, parms, basetype_path, flags)
   /* Declare external function if necessary. */
   assemble_external (function);
 
+#if 1
+  /* Is it a synthesized method that needs to be synthesized?  */
+  if (DECL_ARTIFICIAL (function) && ! flag_no_inline
+      && ! DECL_INITIAL (function)
+      /* Kludge: don't synthesize for default args.  */
+      && current_function_decl)
+    synthesize_method (function);
+#endif
+
+  if (pedantic && DECL_THIS_INLINE (function) && ! DECL_ARTIFICIAL (function)
+       && ! DECL_INITIAL (function) && ! DECL_PENDING_INLINE_INFO (function))
+    cp_pedwarn ("inline function `%#D' called before definition", function);
+
   fntype = TREE_TYPE (function);
   if (TREE_CODE (fntype) == POINTER_TYPE)
     fntype = TREE_TYPE (fntype);
@@ -2566,6 +2636,10 @@ build_method_call (instance, name, parms, basetype_path, flags)
     }
 #endif
 
+  if (parms == error_mark_node
+      || (parms && TREE_CHAIN (parms) == error_mark_node))
+    return error_mark_node;
+
   if (need_vtbl == needed)
     {
       function = build_vfn_ref (&TREE_VALUE (parms), instance,
@@ -2583,14 +2657,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
     if (TREE_CODE (function) == FUNCTION_DECL)
       {
 	is_constructor = DECL_CONSTRUCTOR_P (function);
-	if (DECL_INLINE (function))
-	  function = build1 (ADDR_EXPR, build_pointer_type (fntype), function);
-	else
-	  {
-	    assemble_external (function);
-	    TREE_USED (function) = 1;
-	    function = default_conversion (function);
-	  }
+	TREE_USED (function) = 1;
+	function = default_conversion (function);
       }
     else
       {
@@ -2602,9 +2670,8 @@ build_method_call (instance, name, parms, basetype_path, flags)
 
     TREE_TYPE (result) = value_type;
     TREE_SIDE_EFFECTS (result) = 1;
-    TREE_RAISES (result)
-      = TYPE_RAISES_EXCEPTIONS (fntype) || (parms && TREE_RAISES (parms));
     TREE_HAS_CONSTRUCTOR (result) = is_constructor;
+    result = convert_from_reference (result);
     return result;
   }
 }
@@ -2664,7 +2731,11 @@ build_overload_call_real (fnname, parms, flags, final_cp, buildxxx)
 	    final_cp->h.code = EVIL_CODE;
 	  return error_mark_node;
 	}
-      if (TREE_CODE (t) == ARRAY_TYPE || TREE_CODE (t) == OFFSET_TYPE)
+      if (TREE_CODE (t) == OFFSET_TYPE)
+#if 0
+      /* This breaks reference-to-array parameters.  */
+	  || TREE_CODE (t) == ARRAY_TYPE
+#endif
 	{
 	  /* Perform the conversion from ARRAY_TYPE to POINTER_TYPE in place.
 	     Also convert OFFSET_TYPE entities to their normal selves.
@@ -2738,7 +2809,7 @@ build_overload_call_real (fnname, parms, flags, final_cp, buildxxx)
     {
       candidates
 	= (struct candidate *)alloca ((length+1) * sizeof (struct candidate));
-      bzero (candidates, (length + 1) * sizeof (struct candidate));
+      bzero ((char *) candidates, (length + 1) * sizeof (struct candidate));
     }
 
   cp = candidates;
