@@ -1,5 +1,5 @@
 /* Convert RTL to assembler code and output it, for GNU compiler.
-   Copyright (C) 1987, 88, 89, 92, 93, 1994 Free Software Foundation, Inc.
+   Copyright (C) 1987, 88, 89, 92, 93, 94, 1995 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -15,7 +15,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 
 /* This is the final pass of the compiler.
@@ -121,6 +122,12 @@ static rtx debug_insn = 0;
 /* Line number of last NOTE.  */
 static int last_linenum;
 
+/* Highest line number in current block.  */
+static int high_block_linenum;
+
+/* Likewise for function.  */
+static int high_function_linenum;
+
 /* Filename of last NOTE.  */
 static char *last_filename;
 
@@ -270,6 +277,7 @@ static int add_bb_string	PROTO((char *, int));
 static void output_source_line	PROTO((FILE *, rtx));
 static rtx walk_alter_subreg	PROTO((rtx));
 static int alter_cond		PROTO((rtx));
+static void output_asm_name	PROTO((void));
 static void output_operand	PROTO((rtx, int));
 static void leaf_renumber_regs	PROTO((rtx));
 
@@ -808,6 +816,9 @@ shorten_branches (first)
 	      something_changed = 1;
 	    }
 	}
+      /* For a non-optimizing compile, do only a single pass.  */
+      if (!optimize)
+	break;
     }
 #endif /* HAVE_ATTR_length */
 }
@@ -824,9 +835,13 @@ asm_insn_count (body)
   char *template;
   int count = 1;
 
-  for (template = decode_asm_operands (body, NULL_PTR, NULL_PTR,
-				       NULL_PTR, NULL_PTR);
-       *template; template++)
+  if (GET_CODE (body) == ASM_INPUT)
+    template = XSTR (body, 0);
+  else
+    template = decode_asm_operands (body, NULL_PTR, NULL_PTR,
+				    NULL_PTR, NULL_PTR);
+
+  for ( ; *template; template++)
     if (IS_ASM_LOGICAL_LINE_SEPARATOR(*template) || *template == '\n')
       count++;
 
@@ -873,19 +888,19 @@ final_start_function (first, file, optimize)
      in the last statement of the preceding function.  */
   if (NOTE_LINE_NUMBER (first) != NOTE_INSN_DELETED)
     {
+      last_linenum = high_block_linenum = high_function_linenum
+	= NOTE_LINE_NUMBER (first);
+
       if (write_symbols == SDB_DEBUG)
 	/* For sdb, let's not, but say we did.
 	   We need to set last_linenum for sdbout_function_begin,
 	   but we can't have an actual line number before the .bf symbol.
 	   (sdb_begin_function_line is not set,
 	   and other compilers don't do it.)  */
-	last_linenum = NOTE_LINE_NUMBER (first);
+	;
 #ifdef XCOFF_DEBUGGING_INFO
       else if (write_symbols == XCOFF_DEBUG)
-	{
-	  last_linenum = NOTE_LINE_NUMBER (first);
-	  xcoffout_output_first_source_line (file, last_linenum);
-	}
+	xcoffout_output_first_source_line (file, last_linenum);
 #endif	  
       else
 	output_source_line (file, first);
@@ -1029,7 +1044,7 @@ final_end_function (first, file, optimize)
 
 #ifdef SDB_DEBUGGING_INFO
   if (write_symbols == SDB_DEBUG)
-    sdbout_end_function (last_linenum);
+    sdbout_end_function (high_function_linenum);
 #endif
 
 #ifdef DWARF_DEBUGGING_INFO
@@ -1039,7 +1054,7 @@ final_end_function (first, file, optimize)
 
 #ifdef XCOFF_DEBUGGING_INFO
   if (write_symbols == XCOFF_DEBUG)
-    xcoffout_end_function (file, last_linenum);
+    xcoffout_end_function (file, high_function_linenum);
 #endif
 
 #ifdef FUNCTION_EPILOGUE
@@ -1340,6 +1355,8 @@ final_scan_insn (insn, file, optimize, prescan, nopeepholes)
 	    }
 	  pending_blocks[block_depth++] = next_block_index;
 
+	  high_block_linenum = last_linenum;
+
 	  /* Output debugging info about the symbol-block beginning.  */
 
 #ifdef SDB_DEBUGGING_INFO
@@ -1377,7 +1394,8 @@ final_scan_insn (insn, file, optimize, prescan, nopeepholes)
 
 #ifdef XCOFF_DEBUGGING_INFO
 	  if (write_symbols == XCOFF_DEBUG && block_depth >= 0)
-	    xcoffout_end_block (file, last_linenum, pending_blocks[block_depth]);
+	    xcoffout_end_block (file, high_block_linenum,
+				pending_blocks[block_depth]);
 #endif
 #ifdef DBX_DEBUGGING_INFO
 	  if (write_symbols == DBX_DEBUG && block_depth >= 0)
@@ -1386,7 +1404,8 @@ final_scan_insn (insn, file, optimize, prescan, nopeepholes)
 #endif
 #ifdef SDB_DEBUGGING_INFO
 	  if (write_symbols == SDB_DEBUG && block_depth >= 0)
-	    sdbout_end_block (file, last_linenum, pending_blocks[block_depth]);
+	    sdbout_end_block (file, high_block_linenum,
+			      pending_blocks[block_depth]);
 #endif
 #ifdef DWARF_DEBUGGING_INFO
 	  if (write_symbols == DWARF_DEBUG && block_depth >= 1)
@@ -1497,7 +1516,7 @@ final_scan_insn (insn, file, optimize, prescan, nopeepholes)
 					    / BITS_PER_UNIT));
 #endif /* READONLY_DATA_SECTION */
 #else /* JUMP_TABLES_IN_TEXT_SECTION */
-	      text_section ();
+	      function_section (current_function_decl);
 #endif /* JUMP_TABLES_IN_TEXT_SECTION */
 #ifdef ASM_OUTPUT_CASE_LABEL
 	      ASM_OUTPUT_CASE_LABEL (file, "L", CODE_LABEL_NUMBER (insn),
@@ -1586,7 +1605,7 @@ final_scan_insn (insn, file, optimize, prescan, nopeepholes)
 				 insn);
 #endif
 
-	    text_section ();
+	    function_section (current_function_decl);
 
 	    break;
 	  }
@@ -2014,6 +2033,8 @@ output_source_line (file, insn)
 
   last_filename = filename;
   last_linenum = NOTE_LINE_NUMBER (insn);
+  high_block_linenum = MAX (last_linenum, high_block_linenum);
+  high_function_linenum = MAX (last_linenum, high_function_linenum);
 
   if (write_symbols != NO_DEBUG)
     {
@@ -2073,10 +2094,9 @@ alter_subreg (x)
   else if (GET_CODE (y) == MEM)
     {
       register int offset = SUBREG_WORD (x) * UNITS_PER_WORD;
-#if BYTES_BIG_ENDIAN
-      offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (x)))
-		 - MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (y))));
-#endif
+      if (BYTES_BIG_ENDIAN)
+	offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (x)))
+		   - MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (y))));
       PUT_CODE (x, MEM);
       MEM_VOLATILE_P (x) = MEM_VOLATILE_P (y);
       XEXP (x, 0) = plus_constant (XEXP (y, 0), offset);
@@ -2298,6 +2318,28 @@ output_operand_lossage (str)
       and print a constant expression for minus the value
       of the operand, with no other punctuation.  */
 
+static void
+output_asm_name ()
+{
+  if (flag_print_asm_name)
+    {
+      /* Annotate the assembly with a comment describing the pattern and
+	 alternative used.  */
+      if (debug_insn)
+	{
+	  register int num = INSN_CODE (debug_insn);
+	  fprintf (asm_out_file, " %s %d %s", 
+		   ASM_COMMENT_START, INSN_UID (debug_insn), insn_name[num]);
+	  if (insn_n_alternatives[num] > 1)
+	    fprintf (asm_out_file, "/%d", which_alternative + 1);
+
+	  /* Clear this so only the first assembler insn
+	     of any rtl insn will get the special comment for -dp.  */
+	  debug_insn = 0;
+	}
+    }
+}
+
 void
 output_asm_insn (template, operands)
      char *template;
@@ -2321,17 +2363,18 @@ output_asm_insn (template, operands)
   while (c = *p++)
     switch (c)
       {
-#ifdef ASM_OUTPUT_OPCODE
       case '\n':
+	output_asm_name ();
 	putc (c, asm_out_file);
+#ifdef ASM_OUTPUT_OPCODE
 	while ((c = *p) == '\t')
 	  {
 	    putc (c, asm_out_file);
 	    p++;
 	  }
 	ASM_OUTPUT_OPCODE (asm_out_file, p);
-	break;
 #endif
+	break;
 
 #ifdef ASSEMBLER_DIALECT
       case '{':
@@ -2444,23 +2487,7 @@ output_asm_insn (template, operands)
 	putc (c, asm_out_file);
       }
 
-  if (flag_print_asm_name)
-    {
-      /* Annotate the assembly with a comment describing the pattern and
-	 alternative used.  */
-      if (debug_insn)
-	{
-	  register int num = INSN_CODE (debug_insn);
-	  fprintf (asm_out_file, " %s %d %s", 
-		   ASM_COMMENT_START, INSN_UID (debug_insn), insn_name[num]);
-	  if (insn_n_alternatives[num] > 1)
-	    fprintf (asm_out_file, "/%d", which_alternative + 1);
-
-	  /* Clear this so only the first assembler insn
-	     of any rtl insn will get the special comment for -dp.  */
-	  debug_insn = 0;
-	}
-    }
+  output_asm_name ();
 
   putc ('\n', asm_out_file);
 }
@@ -2831,27 +2858,60 @@ split_double (value, first, second)
 {
   if (GET_CODE (value) == CONST_INT)
     {
-      /* The rule for using CONST_INT for a wider mode
-	 is that we regard the value as signed.
-	 So sign-extend it.  */
-      rtx high = (INTVAL (value) < 0 ? constm1_rtx : const0_rtx);
-#if WORDS_BIG_ENDIAN
-      *first = high;
-      *second = value;
-#else
-      *first = value;
-      *second = high;
-#endif
+      if (HOST_BITS_PER_WIDE_INT >= (2 * BITS_PER_WORD))
+	{
+	  /* In this case the CONST_INT holds both target words.
+	     Extract the bits from it into two word-sized pieces.  */
+	  rtx low, high;
+	  HOST_WIDE_INT word_mask;
+	  /* Avoid warnings for shift count >= BITS_PER_WORD.  */
+	  int shift_count = BITS_PER_WORD - 1;
+
+	  word_mask = (HOST_WIDE_INT) 1 << shift_count;
+	  word_mask |= word_mask - 1;
+	  low = GEN_INT (INTVAL (value) & word_mask);
+	  high = GEN_INT ((INTVAL (value) >> (shift_count + 1)) & word_mask);
+	  if (WORDS_BIG_ENDIAN)
+	    {
+	      *first = high;
+	      *second = low;
+	    }
+	  else
+	    {
+	      *first = low;
+	      *second = high;
+	    }
+	}
+      else
+	{
+	  /* The rule for using CONST_INT for a wider mode
+	     is that we regard the value as signed.
+	     So sign-extend it.  */
+	  rtx high = (INTVAL (value) < 0 ? constm1_rtx : const0_rtx);
+	  if (WORDS_BIG_ENDIAN)
+	    {
+	      *first = high;
+	      *second = value;
+	    }
+	  else
+	    {
+	      *first = value;
+	      *second = high;
+	    }
+	}
     }
   else if (GET_CODE (value) != CONST_DOUBLE)
     {
-#if WORDS_BIG_ENDIAN
-      *first = const0_rtx;
-      *second = value;
-#else
-      *first = value;
-      *second = const0_rtx;
-#endif
+      if (WORDS_BIG_ENDIAN)
+	{
+	  *first = const0_rtx;
+	  *second = value;
+	}
+      else
+	{
+	  *first = value;
+	  *second = const0_rtx;
+	}
     }
   else if (GET_MODE (value) == VOIDmode
 	   /* This is the old way we did CONST_DOUBLE integers.  */
@@ -2859,13 +2919,16 @@ split_double (value, first, second)
     {
       /* In an integer, the words are defined as most and least significant.
 	 So order them by the target's convention.  */
-#if WORDS_BIG_ENDIAN
-      *first = GEN_INT (CONST_DOUBLE_HIGH (value));
-      *second = GEN_INT (CONST_DOUBLE_LOW (value));
-#else
-      *first = GEN_INT (CONST_DOUBLE_LOW (value));
-      *second = GEN_INT (CONST_DOUBLE_HIGH (value));
-#endif
+      if (WORDS_BIG_ENDIAN)
+	{
+	  *first = GEN_INT (CONST_DOUBLE_HIGH (value));
+	  *second = GEN_INT (CONST_DOUBLE_LOW (value));
+	}
+      else
+	{
+	  *first = GEN_INT (CONST_DOUBLE_LOW (value));
+	  *second = GEN_INT (CONST_DOUBLE_HIGH (value));
+	}
     }
   else
     {
@@ -2887,14 +2950,23 @@ split_double (value, first, second)
 	  && ! flag_pretend_float)
       abort ();
 
-#if defined (HOST_WORDS_BIG_ENDIAN) == WORDS_BIG_ENDIAN
-      /* Host and target agree => no need to swap.  */
-      *first = GEN_INT (CONST_DOUBLE_LOW (value));
-      *second = GEN_INT (CONST_DOUBLE_HIGH (value));
+      if (
+#ifdef HOST_WORDS_BIG_ENDIAN
+	  WORDS_BIG_ENDIAN
 #else
-      *second = GEN_INT (CONST_DOUBLE_LOW (value));
-      *first = GEN_INT (CONST_DOUBLE_HIGH (value));
+	  ! WORDS_BIG_ENDIAN
 #endif
+	  )
+	{
+	  /* Host and target agree => no need to swap.  */
+	  *first = GEN_INT (CONST_DOUBLE_LOW (value));
+	  *second = GEN_INT (CONST_DOUBLE_HIGH (value));
+	}
+      else
+	{
+	  *second = GEN_INT (CONST_DOUBLE_LOW (value));
+	  *first = GEN_INT (CONST_DOUBLE_HIGH (value));
+	}
 #endif /* no REAL_ARITHMETIC */
     }
 }
