@@ -1,7 +1,7 @@
 /* Read and manage MIPS symbol tables from object modules.
-   Source originally from hartzell@boulder.colorado.edu
-   Rewritten by: meissner@osf.org
-   Copyright (C) 1991 Free Software Foundation, Inc.
+   Copyright (C) 1991, 1994 Free Software Foundation, Inc.
+   Contributed by hartzell@boulder.colorado.edu,
+   Rewritten by meissner@osf.org.
 
 This file is part of GNU CC.
 
@@ -17,11 +17,13 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/file.h>
+#include <time.h>
 #include <fcntl.h>
 #include <errno.h>
 #include "config.h"
@@ -273,7 +275,6 @@ extern PTR_T	calloc	__proto((size_t, size_t));
 extern PTR_T	realloc	__proto((PTR_T, size_t));
 extern void	free	__proto((PTR_T));
 #endif
-extern char    *ctime	__proto((time_t *));
 
 extern char *optarg;
 extern int   optind;
@@ -594,7 +595,7 @@ type_to_string (aux_ptr, index)
       break;
 
     case bt_Indirect:		/* forward or unnamed typedef */
-      strcpy (p1, "forward/unamed typedef");
+      strcpy (p1, "forward/unnamed typedef");
       break;
 
     case bt_FixedDec:		/* Fixed Decimal */
