@@ -1,22 +1,21 @@
 /* GNU Objective C Runtime selector related functions
-   Copyright (C) 1993 Free Software Foundation, Inc.
-
-Author: Kresten Krab Thorup
+   Copyright (C) 1993, 1995 Free Software Foundation, Inc.
+   Contributed by Kresten Krab Thorup
 
 This file is part of GNU CC.
 
 GNU CC is free software; you can redistribute it and/or modify it under the
-   terms of the GNU General Public License as published by the Free Software
-   Foundation; either version 2, or (at your option) any later version.
+terms of the GNU General Public License as published by the Free Software
+Foundation; either version 2, or (at your option) any later version.
 
 GNU CC is distributed in the hope that it will be useful, but WITHOUT ANY
-   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
-   details.
+WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
+details.
 
 You should have received a copy of the GNU General Public License along with
-   GNU CC; see the file COPYING.  If not, write to the Free Software
-   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+GNU CC; see the file COPYING.  If not, write to the Free Software
+Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 /* As a special exception, if you link this library with files compiled with
    GCC to produce an executable, this does not cause the resulting executable
@@ -26,8 +25,9 @@ You should have received a copy of the GNU General Public License along with
 
 #include "runtime.h"
 #include "objc/sarray.h"
+#include "encoding.h"
 
-/* Initial selector hash table size. Value doesnt matter much */
+/* Initial selector hash table size. Value doesn't matter much */
 #define SELECTOR_HASH_SIZE 128
 
 /* Tables mapping selector names to uid and opposite */
@@ -53,7 +53,7 @@ void __objc_init_selector_tables()
 /* This routine is given a class and records all of the methods in its class
    structure in the record table.  */
 void
-__objc_register_selectors_from_class (Class* class)
+__objc_register_selectors_from_class (Class class)
 {
   MethodList_t method_list;
 
@@ -87,6 +87,34 @@ register_selectors_from_list (MethodList_t method_list)
     }
 }
 
+
+/* Returns YES iff t1 and t2 have same method types, but we ignore
+   the argframe layout */
+BOOL
+sel_types_match (const char* t1, const char* t2)
+{
+  if (!t1 || !t2)
+    return NO;
+  while (*t1 && *t2)
+    {
+      if (*t1 == '+') t1++;
+      if (*t2 == '+') t2++;
+      while (isdigit(*t1)) t1++;
+      while (isdigit(*t2)) t2++;
+      /* xxx Remove these next two lines when qualifiers are put in
+	 all selectors, not just Protocol selectors. */
+      t1 = objc_skip_type_qualifiers(t1);
+      t2 = objc_skip_type_qualifiers(t2);
+      if (!*t1 && !*t2)
+	return YES;
+      if (*t1 != *t2)
+	return NO;
+      t1++;
+      t2++;
+    }
+  return NO;
+}
+
 /* return selector representing name */
 SEL
 sel_get_typed_uid (const char *name, const char *types)
@@ -109,7 +137,7 @@ sel_get_typed_uid (const char *name, const char *types)
 	      return s;
 	    }
 	}
-      else if (! strcmp (s->sel_types, types))
+      else if (sel_types_match (s->sel_types, types))
 	{
 	  return s;
 	}
@@ -150,7 +178,7 @@ sel_get_name (SEL selector)
 {
   if ((soffset_decode((sidx)selector->sel_id) > 0)
       && (soffset_decode((sidx)selector->sel_id) <= __objc_selector_max_index))
-    return sarray_get (__objc_selector_array, (sidx) selector->sel_id);
+    return sarray_get (__objc_selector_names, (sidx) selector->sel_id);
   else
     return 0;
 }
@@ -205,7 +233,7 @@ __sel_register_typed_name (const char *name, const char *types,
 		    return s;
 		}
 	    }
-	  else if (strcmp (s->sel_types, types))
+	  else if (!strcmp (s->sel_types, types))
 	    {
 	      if (orig)
 		{
