@@ -15,7 +15,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
 
 
 #include "config.h"
@@ -84,7 +85,7 @@ init_recog ()
 
 /* Try recognizing the instruction INSN,
    and return the code number that results.
-   Remeber the code so that repeated calls do not
+   Remember the code so that repeated calls do not
    need to spend the time for actual rerecognition.
 
    This function is the normal interface to instruction recognition.
@@ -423,11 +424,10 @@ validate_replace_rtx_1 (loc, from, to, object)
 	  enum machine_mode mode = GET_MODE (x);
 	  rtx new;
 
-#if BYTES_BIG_ENDIAN
-	  offset += (MIN (UNITS_PER_WORD,
-			  GET_MODE_SIZE (GET_MODE (SUBREG_REG (x))))
-		     - MIN (UNITS_PER_WORD, GET_MODE_SIZE (mode)));
-#endif
+	  if (BYTES_BIG_ENDIAN)
+	    offset += (MIN (UNITS_PER_WORD,
+			    GET_MODE_SIZE (GET_MODE (SUBREG_REG (x))))
+		       - MIN (UNITS_PER_WORD, GET_MODE_SIZE (mode)));
 
 	  new = gen_rtx (MEM, mode, plus_constant (XEXP (to, 0), offset));
 	  MEM_VOLATILE_P (new) = MEM_VOLATILE_P (to);
@@ -474,10 +474,9 @@ validate_replace_rtx_1 (loc, from, to, object)
 
 		  /* If the bytes and bits are counted differently, we
 		     must adjust the offset.  */
-#if BYTES_BIG_ENDIAN != BITS_BIG_ENDIAN
-	      offset = (GET_MODE_SIZE (is_mode) - GET_MODE_SIZE (wanted_mode)
-			- offset);
-#endif
+	      if (BYTES_BIG_ENDIAN != BITS_BIG_ENDIAN)
+		offset = (GET_MODE_SIZE (is_mode) - GET_MODE_SIZE (wanted_mode)
+			  - offset);
 
 	      pos %= GET_MODE_BITSIZE (wanted_mode);
 
@@ -1083,10 +1082,9 @@ indirect_operand (op, mode)
       register int offset = SUBREG_WORD (op) * UNITS_PER_WORD;
       rtx inner = SUBREG_REG (op);
 
-#if BYTES_BIG_ENDIAN
-      offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (op)))
-		 - MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (inner))));
-#endif
+      if (BYTES_BIG_ENDIAN)
+	offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (op)))
+		   - MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (inner))));
 
       if (mode != VOIDmode && GET_MODE (op) != mode)
 	return 0;
@@ -1718,9 +1716,9 @@ constrain_operands (insn_code_num, strict)
 
 	      case 'p':
 		/* p is used for address_operands.  When we are called by
-		   gen_input_reload, no one will have checked that the
-		   address is strictly valid, i.e., that all pseudos
-		   requiring hard regs have gotten them.  */
+		   gen_reload, no one will have checked that the address is
+		   strictly valid, i.e., that all pseudos requiring hard regs
+		   have gotten them.  */
 		if (strict <= 0
 		    || (strict_memory_address_p
 			(insn_operand_mode[insn_code_num][opno], op)))
@@ -1786,12 +1784,14 @@ constrain_operands (insn_code_num, strict)
 		break;
 
 	      case 'E':
+#ifndef REAL_ARITHMETIC
 		/* Match any CONST_DOUBLE, but only if
 		   we can examine the bits of it reliably.  */
 		if ((HOST_FLOAT_FORMAT != TARGET_FLOAT_FORMAT
 		     || HOST_BITS_PER_WIDE_INT != BITS_PER_WORD)
 		    && GET_MODE (op) != VOIDmode && ! flag_pretend_float)
 		  break;
+#endif
 		if (GET_CODE (op) == CONST_DOUBLE)
 		  win = 1;
 		break;
